"""
Project
-------
"""

from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import numpy as np
import scipy.sparse as sp  # type: ignore
from anndata import AnnData  # type: ignore

import metacells.parameters as pr
import metacells.utilities as ut

__all__ = [
    "compute_projection_weights",
    "compute_projected_fractions",
    "convey_atlas_to_query",
]


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def compute_projection_weights(
    *,
    adata: AnnData,
    qdata: AnnData,
    from_atlas_layer: str = "corrected_fraction",
    from_query_layer: str = "corrected_fraction",
    to_query_layer: str = "projected_fraction",
    log_data: bool = pr.project_log_data,
    fold_regularization: float = pr.project_fold_regularization,
    min_significant_gene_umis: float = pr.project_min_significant_gene_umis,
    max_consistency_fold_factor: float = pr.project_max_consistency_fold_factor,
    candidates_count: int = pr.project_candidates_count,
    min_candidates_fraction: float = pr.project_min_candidates_fraction,
    min_usage_weight: float = pr.project_min_usage_weight,
    second_anchor_indices: Optional[List[int]] = None,
    reproducible: bool,
) -> ut.CompressedMatrix:
    """
    Compute the weights and results of projecting a query onto an atlas.

    **Input**

    Annotated query ``qdata`` and atlas ``adata``, where the observations are cells and the variables are genes. The
    atlas should contain ``from_atlas_layer`` (default: {from_atlas_layer}) containing gene fractions, and the query
    should similarly contain ``from_query_layer`` (default: {from_query_layer}) containing gene fractions.

    **Returns**

    A CSR matrix whose rows are query metacells and columns are atlas metacells, where each entry is the weight of the
    atlas metacell in the projection of the query metacells. The sum of weights in each row (that is, for a single query
    metacell) is 1. The weighted sum of the atlas metacells using these weights is the "projected" image of the query
    metacell onto the atlas.

    In addition, sets the following annotations in ``qdata``:

    Observation (Cell) Annotations
        ``similar``
            A boolean mask indicating whether the query metacell is similar to its projection onto the atlas. If
            ``False`` the metacells is said to be "dissimilar", which may indicate the query contains cell states that
            do not appear in the atlas.

    Observation-Variable (Cell-Gene) Annotations
        ``to_query_layer`` (default: {to_query_layer})
            A matrix of gene fractions describing the "projected" image of the query metacell onto the atlas. This
            projection is a weighted average of some atlas metacells (using the computed weights returned by this
            function).

    **Computation Parameters**

    0. All fold computations (log2 of the ratio between gene fractions) use the ``fold_regularization`` (default:
       {fold_regularization}).

    For each query metacell:

    1. Correlate the metacell with all the atlas metacells, and pick the highest-correlated one as the "anchor".
       If ``second_anchor_indices`` is not ``None``, then the ``qdata`` must contain only a single query metacell, and
       is expected to contain a ``projected`` per-observation-per-variable matrix containing the projected image of this
       query metacell on the atlas using a single anchor. The code will compute the residual of the query and the atlas
       relative to this projection and pick a second atlas anchor whose residuals are the most correlated to the query
       metacell's residuals. If ``reproducible``, a slower (still parallel) but reproducible algorithm will be used.

    2. Consider (for each anchor) the ``candidates_count`` (default: {candidates_count}) candidate metacells with the
       highest correlation with the query metacell.

    3. Keep as candidates only atlas metacells whose maximal gene fold factor compared to the anchor(s) is at most
       ``max_consistency_fold_factor`` (default: {max_consistency_fold_factor}). Keep at least
       ``min_candidates_fraction`` (default: {min_candidates_fraction}) of the original candidates even if they are less
       consistent. For this computation, Ignore the fold factors of genes whose sum of UMIs in the anchor(s) and the
       candidate metacells is less than ``min_significant_gene_umis`` (default: {min_significant_gene_umis}).

    4. Compute the non-negative weights (with a sum of 1) of the selected candidates that give the best projection of
       the query metacells onto the atlas. If ``log_data`` (default: {log_data}), try to fit the log (base 2) of the
       fractions, otherwise, try to fit the fractions themselves. Since the algorithm for computing these weights rarely
       produces an exact 0 weight, reduce all weights less than the ``min_usage_weight`` (default: {min_usage_weight})
       to zero. If ``second_anchor_indices`` is not ``None``, it is set to the list of indices of the used atlas
       metacells candidates correlated with the second anchor.
    """
    prepared_arguments = _project_query_atlas_data_arguments(
        adata=adata,
        qdata=qdata,
        from_atlas_layer=from_atlas_layer,
        from_query_layer=from_query_layer,
        to_query_layer=to_query_layer,
        log_data=log_data,
        fold_regularization=fold_regularization,
        min_significant_gene_umis=min_significant_gene_umis,
        max_consistency_fold_factor=max_consistency_fold_factor,
        candidates_count=candidates_count,
        min_candidates_fraction=min_candidates_fraction,
        min_usage_weight=min_usage_weight,
        second_anchor_indices=second_anchor_indices,
        reproducible=reproducible,
    )

    @ut.timed_call()
    def _project_single(query_metacell_index: int) -> Tuple[ut.NumpyVector, ut.NumpyVector]:
        return _project_single_metacell(
            query_metacell_index=query_metacell_index,
            **prepared_arguments,
        )

    if ut.is_main_process():
        results = ut.parallel_map(_project_single, qdata.n_obs)
    else:
        results = [_project_single(query_metacell_index) for query_metacell_index in range(qdata.n_obs)]

    indices = np.concatenate([result[0] for result in results], dtype="int32")
    data = np.concatenate([result[1] for result in results], dtype="float32")

    atlas_used_sizes = [len(result[0]) for result in results]
    atlas_used_sizes.insert(0, 0)
    indptr = np.cumsum(np.array(atlas_used_sizes))

    return sp.csr_matrix((data, indices, indptr), shape=(qdata.n_obs, adata.n_obs))


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def compute_projected_fractions(
    *,
    adata: AnnData,
    qdata: AnnData,
    from_atlas_layer: str = "corrected_fraction",
    to_query_layer: str = "projected_fraction",
    log_data: bool = pr.project_log_data,
    fold_regularization: float = pr.project_fold_regularization,
    weights: ut.ProperMatrix,
) -> None:
    """
    Compute the projected image of a query on an atlas.

    **Input**

    Annotated query ``qdata`` and atlas ``adata``, where the observations are cells and the variables are genes. The
    atlas should contain ``from_atlas_layer`` (default: {from_atlas_layer}) containing gene fractions.

    **Returns**

    Sets ``to_query_layer`` (default: {to_query_layer}) in the query containing the gene fractions of the projection of
    the atlas fractions using the ``weights`` matrix.

    .. note::

        It is important to use the same ``log_data`` value as that given to ``compute_projection_weights`` to compute
        the weights (default: {log_data}).
    """
    assert fold_regularization > 0

    atlas_fractions = ut.get_vo_proper(adata, from_atlas_layer, layout="row_major")

    if log_data:
        atlas_log_fractions = ut.to_numpy_matrix(atlas_fractions, copy=True)
        atlas_log_fractions += fold_regularization
        atlas_log_fractions = np.log2(atlas_log_fractions, out=atlas_log_fractions)
        projected_fractions = weights @ atlas_log_fractions
        projected_fractions = np.power(2.0, projected_fractions, out=projected_fractions)
        projected_fractions -= fold_regularization

    else:
        projected_fractions = weights @ atlas_fractions  # type: ignore

    assert projected_fractions.shape == qdata.shape
    projected_fractions = ut.fraction_by(projected_fractions, by="row").astype("float32")  # type: ignore
    ut.set_vo_data(qdata, to_query_layer, sp.csr_matrix(projected_fractions))


def _project_query_atlas_data_arguments(
    adata: AnnData,
    qdata: AnnData,
    from_atlas_layer: str,
    from_query_layer: str,
    to_query_layer: str,
    log_data: bool,
    fold_regularization: float,
    min_significant_gene_umis: float,
    max_consistency_fold_factor: float,
    candidates_count: int,
    min_candidates_fraction: float,
    min_usage_weight: float,
    second_anchor_indices: Optional[List[int]],
    reproducible: bool,
) -> Dict[str, Any]:
    assert fold_regularization > 0
    assert candidates_count > 0
    assert 0 <= min_candidates_fraction <= 1.0
    assert min_usage_weight >= 0
    assert max_consistency_fold_factor >= 0
    assert np.all(adata.var_names == qdata.var_names)

    atlas_umis = ut.get_vo_proper(adata, "total_umis", layout="row_major")

    atlas_fractions = ut.get_vo_proper(adata, from_atlas_layer, layout="row_major")
    query_fractions = ut.get_vo_proper(qdata, from_query_layer, layout="row_major")

    atlas_fractions = ut.to_numpy_matrix(atlas_fractions, copy=True)
    query_fractions = ut.to_numpy_matrix(query_fractions, copy=True)

    if second_anchor_indices is not None:
        assert qdata.n_obs == 1
        query_single_fractions = ut.to_numpy_vector(ut.get_vo_proper(qdata, to_query_layer))

        query_residual_fractions = query_fractions - query_single_fractions[np.newaxis, :]
        query_residual_fractions[query_residual_fractions < 0] = 0

        atlas_residual_fractions = atlas_fractions - ut.to_numpy_vector(query_residual_fractions)[np.newaxis, :]
        atlas_residual_fractions[atlas_residual_fractions < 0] = 0

        if log_data:
            atlas_residual_fractions += fold_regularization
            query_residual_fractions += fold_regularization

            atlas_residual_data = np.log2(atlas_residual_fractions, out=atlas_residual_fractions)
            query_residual_data = np.log2(query_residual_fractions, out=query_residual_fractions)

        else:
            atlas_residual_data = atlas_residual_fractions
            query_residual_data = query_residual_fractions

        query_atlas_corr_residual: Optional[ut.NumpyMatrix] = ut.cross_corrcoef_rows(
            query_residual_data, atlas_residual_data, reproducible=reproducible
        )

    else:
        query_atlas_corr_residual = None

    atlas_log_fractions = atlas_fractions + fold_regularization
    atlas_log_fractions = np.log2(atlas_log_fractions, out=atlas_log_fractions)

    query_log_fractions = query_fractions + fold_regularization
    query_log_fractions = np.log2(query_log_fractions, out=query_log_fractions)

    if log_data:
        atlas_data = atlas_log_fractions
        query_data = query_log_fractions

    else:
        atlas_data = atlas_fractions
        query_data = query_fractions

    query_atlas_corr = ut.cross_corrcoef_rows(query_log_fractions, atlas_log_fractions, reproducible=reproducible)

    return {
        "atlas_umis": atlas_umis,
        "query_atlas_corr": query_atlas_corr,
        "atlas_data": atlas_data,
        "atlas_log_fractions": atlas_log_fractions,
        "query_data": query_data,
        "candidates_count": candidates_count,
        "min_candidates_fraction": min_candidates_fraction,
        "min_significant_gene_umis": min_significant_gene_umis,
        "min_usage_weight": min_usage_weight,
        "max_consistency_fold_factor": max_consistency_fold_factor,
        "second_anchor_indices": second_anchor_indices,
        "query_atlas_corr_residual": query_atlas_corr_residual,
    }


@ut.logged()
def _project_single_metacell(  # pylint: disable=too-many-statements,too-many-branches
    *,
    query_metacell_index: int,
    atlas_umis: ut.Matrix,
    query_atlas_corr: ut.NumpyMatrix,
    atlas_data: ut.NumpyMatrix,
    atlas_log_fractions: ut.NumpyMatrix,
    query_data: ut.NumpyMatrix,
    candidates_count: int,
    min_candidates_fraction: float,
    min_significant_gene_umis: float,
    min_usage_weight: float,
    max_consistency_fold_factor: float,
    second_anchor_indices: Optional[List[int]],
    query_atlas_corr_residual: Optional[ut.NumpyMatrix],
) -> Tuple[ut.NumpyVector, ut.NumpyVector]:
    query_metacell_data = query_data[query_metacell_index, :]

    query_metacell_atlas_correlations = query_atlas_corr[query_metacell_index, :]
    query_metacell_atlas_order = np.argsort(-query_metacell_atlas_correlations)

    atlas_anchor_index = query_metacell_atlas_order[0]
    ut.log_calc("atlas_anchor_index", atlas_anchor_index)
    atlas_anchor_log_fractions = atlas_log_fractions[atlas_anchor_index, :]
    atlas_anchor_umis = ut.to_numpy_vector(atlas_umis[atlas_anchor_index, :])

    atlas_candidates_consistency = [0.0]
    atlas_candidates_indices = [atlas_anchor_index]

    position = 1
    while len(atlas_candidates_indices) < candidates_count and position < len(query_metacell_atlas_order):
        atlas_metacell_index = query_metacell_atlas_order[position]
        position += 1
        atlas_metacell_log_fractions = atlas_log_fractions[atlas_metacell_index, :]
        atlas_metacell_consistency_fold_factors = np.abs(atlas_metacell_log_fractions - atlas_anchor_log_fractions)
        atlas_metacell_umis = ut.to_numpy_vector(atlas_umis[atlas_metacell_index, :])
        atlas_metacell_significant_genes_mask = atlas_metacell_umis + atlas_anchor_umis >= min_significant_gene_umis
        if np.any(atlas_metacell_significant_genes_mask):
            atlas_metacell_consistency = np.max(
                atlas_metacell_consistency_fold_factors[atlas_metacell_significant_genes_mask]
            )
            atlas_candidates_consistency.append(atlas_metacell_consistency)
            atlas_candidates_indices.append(atlas_metacell_index)

    sorted_locations = list(np.argsort(np.array(atlas_candidates_consistency)))
    min_candidates_count = candidates_count * min_candidates_fraction

    while (
        len(sorted_locations) > min_candidates_count
        and atlas_candidates_consistency[sorted_locations[-1]] > max_consistency_fold_factor
    ):
        sorted_locations.pop()

    atlas_candidate_indices_set = set([atlas_anchor_index])
    for location in sorted_locations:
        atlas_metacell_index = atlas_candidates_indices[location]
        atlas_candidate_indices_set.add(atlas_metacell_index)

    ut.log_calc("atlas_candidates", len(atlas_candidate_indices_set))

    if query_atlas_corr_residual is None:
        atlas_candidate_indices = np.array(sorted(atlas_candidate_indices_set))
    else:
        query_metacell_atlas_residual_correlations = query_atlas_corr_residual[query_metacell_index, :]
        query_metacell_atlas_residual_order = np.argsort(-query_metacell_atlas_residual_correlations)

        atlas_secondary_anchor_index = query_metacell_atlas_residual_order[0]
        ut.log_calc("atlas_secondary_anchor_index", atlas_secondary_anchor_index)
        atlas_secondary_anchor_log_fractions = atlas_log_fractions[atlas_anchor_index, :]
        atlas_secondary_anchor_umis = ut.to_numpy_vector(atlas_umis[atlas_secondary_anchor_index, :])

        atlas_secondary_candidates_consistency = [0.0]
        atlas_secondary_candidates_indices = [atlas_secondary_anchor_index]

        position = 1
        while len(atlas_secondary_candidates_indices) < candidates_count and position < len(
            query_metacell_atlas_residual_order
        ):
            atlas_metacell_index = query_metacell_atlas_residual_order[position]
            position += 1
            atlas_metacell_log_fractions = atlas_log_fractions[atlas_metacell_index, :]
            atlas_metacell_consistency_fold_factors = np.abs(
                atlas_metacell_log_fractions - atlas_secondary_anchor_log_fractions
            )
            atlas_metacell_umis = ut.to_numpy_vector(atlas_umis[atlas_metacell_index, :])
            atlas_metacell_significant_genes_mask = (
                atlas_metacell_umis + atlas_secondary_anchor_umis >= min_significant_gene_umis
            )
            if np.any(atlas_metacell_significant_genes_mask):
                atlas_metacell_consistency = np.max(
                    atlas_metacell_consistency_fold_factors[atlas_metacell_significant_genes_mask]
                )
                atlas_secondary_candidates_consistency.append(atlas_metacell_consistency)
                atlas_secondary_candidates_indices.append(atlas_metacell_index)

        sorted_secondary_locations = list(np.argsort(np.array(atlas_secondary_candidates_consistency)))

        while (
            len(sorted_secondary_locations) > min_candidates_count
            and atlas_secondary_candidates_consistency[sorted_secondary_locations[-1]] > max_consistency_fold_factor
        ):
            sorted_secondary_locations.pop()

        atlas_secondary_candidate_indices_set = set([atlas_secondary_anchor_index])
        for location in sorted_secondary_locations:
            atlas_metacell_index = atlas_secondary_candidates_indices[location]
            atlas_secondary_candidate_indices_set.add(atlas_metacell_index)

        ut.log_calc("atlas_secondary_candidates", len(atlas_candidate_indices_set))

        atlas_candidate_indices = np.array(sorted(atlas_candidate_indices_set | atlas_secondary_candidate_indices_set))

    atlas_candidates_data = atlas_data[atlas_candidate_indices, :]
    represent_result = ut.represent(query_metacell_data, atlas_candidates_data)
    assert represent_result is not None
    atlas_candidate_weights = represent_result[1]
    atlas_candidate_weights[atlas_candidate_weights < min_usage_weight] = 0
    atlas_candidate_weights[:] /= np.sum(atlas_candidate_weights)

    atlas_used_mask = atlas_candidate_weights > 0

    atlas_used_indices = atlas_candidate_indices[atlas_used_mask].astype("int32")
    ut.log_return("atlas_used_indices", atlas_used_indices)

    if second_anchor_indices is not None:
        for atlas_metacell_index in atlas_used_indices:
            if atlas_metacell_index not in atlas_candidate_indices_set:
                second_anchor_indices.append(atlas_metacell_index)

    atlas_used_weights = atlas_candidate_weights[atlas_used_mask]
    atlas_used_weights = atlas_used_weights.astype("float32")
    ut.log_return("atlas_used_weights", atlas_used_weights)

    return (atlas_used_indices, atlas_used_weights)


@ut.logged()
@ut.timed_call()
def convey_atlas_to_query(
    *,
    adata: AnnData,
    qdata: AnnData,
    weights: ut.ProperMatrix,
    property_name: str,
    formatter: Optional[Callable[[Any], Any]] = None,
    to_property_name: Optional[str] = None,
    method: Callable[[ut.Vector, ut.Vector], Any] = ut.highest_weight,
) -> None:
    """
    Convey the value of a property from per-observation atlas data to per-observation query data.

    The input annotated ``adata`` is expected to contain a per-observation (cell) annotation named ``property_name``.
    Given the ``weights`` matrix, where each row specifies the weights of the atlas metacells used to project a single
    query metacell, this will generate a new per-observation (group) annotation in ``qdata``, named ``to_property_name``
    (by default, the same as ``property_name``), containing the aggregated value of the property of all the observations
    (cells) that belong to the group.

    The aggregation method (by default, :py:func:`metacells.utilities.computation.highest_weight`) is any function
    taking two array, weights and values, and returning a single value.
    """
    if to_property_name is None:
        to_property_name = property_name

    property_of_atlas_metacells = ut.get_o_numpy(adata, property_name, formatter=formatter)
    property_of_query_metacells = []
    for query_metacell_index in range(qdata.n_obs):
        atlas_metacell_weights = ut.to_numpy_vector(weights[query_metacell_index, :])
        used_atlas_metacells_mask = atlas_metacell_weights > 0
        assert np.any(used_atlas_metacells_mask)
        used_atlas_metacell_weights = ut.to_numpy_vector(atlas_metacell_weights[used_atlas_metacells_mask])
        used_atlas_metacell_values = property_of_atlas_metacells[used_atlas_metacells_mask]
        property_of_query_metacells.append(method(used_atlas_metacell_weights, used_atlas_metacell_values))

    query_property_data = np.array(property_of_query_metacells, dtype=property_of_atlas_metacells.dtype)
    ut.set_o_data(qdata, to_property_name, query_property_data)



"""
Dissolve
--------
"""

from math import floor
from typing import Optional
from typing import Union

import numpy as np
from anndata import AnnData  # type: ignore

import metacells.parameters as pr
import metacells.utilities as ut

__all__ = [
    "dissolve_metacells",
]


@ut.logged()
@ut.expand_doc()
@ut.timed_call()
def dissolve_metacells(
    adata: AnnData,
    what: Union[str, ut.Matrix] = "__x__",
    *,
    candidates: Union[str, ut.Vector] = "candidate",
    deviants: ut.Vector,
    target_metacell_size: int = pr.target_metacell_size,
    min_metacell_size: int = pr.min_metacell_size,
    target_metacell_umis: int = pr.target_metacell_umis,
    cell_umis: Optional[ut.NumpyVector] = pr.cell_umis,
    min_robust_size_factor: float = pr.dissolve_min_robust_size_factor,
    min_convincing_gene_fold_factor: Optional[float] = pr.dissolve_min_convincing_gene_fold_factor,
) -> None:
    """
    Dissolve too-small metacells based on ``what`` (default: {what}) data.

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes, where
    ``what`` is a per-variable-per-observation matrix or the name of a per-variable-per-observation
    annotation containing such a matrix.

    **Returns**

    Sets the following in ``adata``:

    Observation (Cell) Annotations
        ``metacell``
            The integer index of the metacell each cell belongs to. The metacells are in no
            particular order. Cells with no metacell assignment are given a metacell index of
            ``-1``.

        ``dissolved``
           A boolean mask of the cells which were in a dissolved metacell.

    **Computation Parameters**

    0. If ``cell_umis`` is not specified, use the sum of the ``what`` data for each cell.

    1. Mark all ``deviants`` cells "outliers". This can be the name of a per-observation (cell) annotation, or an
       explicit boolean mask of cells, or a or ``None`` if there are no deviant cells to mark.

    2. Any metacell which has less cells than the ``min_metacell_size`` is dissolved into outlier cells.

    3. If ``min_convincing_gene_fold_factor`` is not ``None``, preserve everything else. Otherwise:

    4. We are trying to create metacells of size ``target_metacell_size`` cells and ``target_metacell_umis`` UMIs each.
       Compute the UMIs of the metacells by summing the ``cell_umis``.

    5. Using ``min_robust_size_factor`` (default: {min_robust_size_factor}), any
       metacell whose total size is at least ``target_metacell_size * min_robust_size_factor``
       or whose total UMIs are at least ``target_metacell_umis * min_robust_size_factor`` is preserved.

    6. Using ``min_convincing_gene_fold_factor``, preserve any remaining metacells which have at least one gene whose
       fold factor (log2((actual + 1) / (expected_by_overall_population + 1))) is at least this high.

    6. Dissolve the remaining metacells into outlier cells.
    """
    dissolved_of_cells = np.zeros(adata.n_obs, dtype="bool")

    candidate_of_cells = ut.get_o_numpy(adata, candidates, formatter=ut.groups_description)
    candidate_of_cells = np.copy(candidate_of_cells)

    deviant_of_cells = ut.maybe_o_numpy(adata, deviants, formatter=ut.mask_description)
    if deviant_of_cells is not None:
        deviant_of_cells = deviant_of_cells > 0

    if cell_umis is None:
        cell_umis = ut.get_o_numpy(adata, what, sum=True, formatter=ut.sizes_description).astype("float32")
    else:
        assert cell_umis.dtype == "float32"
    assert isinstance(cell_umis, ut.NumpyVector)

    if deviant_of_cells is not None:
        candidate_of_cells[deviant_of_cells > 0] = -1
    candidate_of_cells = ut.compress_indices(candidate_of_cells)
    candidates_count = np.max(candidate_of_cells) + 1

    data = ut.get_vo_proper(adata, what, layout="column_major")
    fraction_of_genes = ut.fraction_per(data, per="column")

    min_robust_size = int(floor(target_metacell_size * min_robust_size_factor))
    min_robust_umis = int(floor(target_metacell_umis * min_robust_size_factor))
    ut.log_calc("min_robust_size", min_robust_size)
    ut.log_calc("min_robust_umis", min_robust_umis)

    did_dissolve = False
    for candidate_index in range(candidates_count):
        candidate_cell_indices = np.where(candidate_of_cells == candidate_index)[0]
        if not _keep_candidate(
            candidate_index,
            data=data,
            cell_umis=cell_umis,
            fraction_of_genes=fraction_of_genes,
            min_metacell_size=min_metacell_size,
            min_robust_size=min_robust_size,
            min_robust_umis=min_robust_umis,
            min_convincing_gene_fold_factor=min_convincing_gene_fold_factor,
            candidates_count=candidates_count,
            candidate_cell_indices=candidate_cell_indices,
        ):
            dissolved_of_cells[candidate_cell_indices] = True
            candidate_of_cells[candidate_cell_indices] = -1
            did_dissolve = True

    if did_dissolve:
        metacell_of_cells = ut.compress_indices(candidate_of_cells)
    else:
        metacell_of_cells = candidate_of_cells

    ut.set_o_data(adata, "dissolved", dissolved_of_cells)
    ut.set_o_data(adata, "metacell", metacell_of_cells, formatter=ut.groups_description)


def _keep_candidate(
    candidate_index: int,
    *,
    data: ut.ProperMatrix,
    cell_umis: ut.NumpyVector,
    fraction_of_genes: ut.NumpyVector,
    min_metacell_size: int,
    min_robust_size: int,
    min_robust_umis: int,
    min_convincing_gene_fold_factor: Optional[float],
    candidates_count: int,
    candidate_cell_indices: ut.NumpyVector,
) -> bool:
    metacell_size = candidate_cell_indices.size
    metacell_umis = np.sum(cell_umis[candidate_cell_indices])

    if metacell_size < min_metacell_size:
        if ut.logging_calc():
            ut.log_calc(
                f'- candidate: {ut.progress_description(candidates_count, candidate_index, "candidate")} '
                f"size: {metacell_size} "
                f"umis: {metacell_umis:g} "
                f"has: less than minimal size"
            )
        return False

    if metacell_size >= min_robust_size:
        if ut.logging_calc():
            ut.log_calc(
                f'- candidate: {ut.progress_description(candidates_count, candidate_index, "candidate")} '
                f"size: {metacell_size} "
                f"umis: {metacell_umis:g} "
                f"has: robust size"
            )
        return True

    if metacell_umis >= min_robust_umis:
        if ut.logging_calc():
            ut.log_calc(
                f'- candidate: {ut.progress_description(candidates_count, candidate_index, "candidate")} '
                f"size: {metacell_size} "
                f"umis: {metacell_umis:g} "
                f"has: robust umis"
            )
        return True

    if min_convincing_gene_fold_factor is None:
        if ut.logging_calc():
            ut.log_calc(
                f'- candidate: {ut.progress_description(candidates_count, candidate_index, "candidate")} '
                f"size: {metacell_size} "
                f"umis: {metacell_umis:g} "
                f"has: minimal size"
            )
        return True

    genes_count = data.shape[1]
    candidate_data = data[candidate_cell_indices, :]
    candidate_data_of_genes = ut.to_numpy_vector(candidate_data.sum(axis=0))
    assert candidate_data_of_genes.size == genes_count
    candidate_total = np.sum(candidate_data_of_genes)
    candidate_expected_of_genes = fraction_of_genes * candidate_total
    candidate_expected_of_genes += 1
    candidate_data_of_genes += 1
    candidate_data_of_genes /= candidate_expected_of_genes
    np.log2(candidate_data_of_genes, out=candidate_data_of_genes)
    convincing_genes_mask = np.abs(candidate_data_of_genes) >= min_convincing_gene_fold_factor
    keep_candidate = bool(np.any(convincing_genes_mask))

    if ut.logging_calc():
        if keep_candidate:
            ut.log_calc(
                f'- candidate: {ut.progress_description(candidates_count, candidate_index, "candidate")} '
                f"size: {metacell_size} "
                f"umis: {metacell_umis:g} "
                f"has: convincing genes"
            )
        else:
            ut.log_calc(
                f'- candidate: {ut.progress_description(candidates_count, candidate_index, "candidate")} '
                f"size: {metacell_size} "
                f"umis: {metacell_umis:g} "
                f"has: no convincing genes"
            )

    return keep_candidate



"""
Test applying functions to real data.
"""

import logging
from glob import glob
from typing import Any
from typing import Dict
from typing import Tuple

import numpy as np
import scanpy as sc  # type: ignore
import yaml  # type: ignore
from anndata import AnnData  # type: ignore

import metacells as mc

# pylint: disable=missing-function-docstring

np.seterr(all="raise")
mc.ut.setup_logger(level=logging.WARN)
mc.ut.allow_inefficient_layout(False)
mc.ut.set_processors_count(4)

LOADED: Dict[str, Tuple[AnnData, Dict[str, Any]]] = {}


def _load(path: str) -> Tuple[AnnData, Dict[str, Any]]:
    if path in LOADED:
        return LOADED[path]

    with open(path[:-4] + "yaml", encoding="utf8") as file:
        expected = yaml.safe_load(file)
    with mc.ut.timed_step("read"):
        adata = sc.read(path)

    LOADED[path] = (adata, expected)
    mc.ut.set_name(adata, path.split("/")[-1][:-5])
    return adata, expected


def test_find_rare_gene_modules() -> None:
    for path in glob("../metacells-test-data/*.h5ad"):
        adata, expected = _load(path)

        mc.pl.mark_lateral_genes(adata, **expected.get("mark_lateral_genes", {}))
        mc.tl.find_rare_gene_modules(  #
            adata,
            reproducible=True,
            **expected.get("find_rare_gene_modules", {}),
        )

        actual_rare_gene_modules = []
        rare_gene_modules = mc.ut.get_v_numpy(adata, "rare_gene_module")
        max_gene_module = np.max(rare_gene_modules)
        for module_index in range(max_gene_module + 1):
            actual_rare_gene_modules.append(sorted(adata.var_names[rare_gene_modules == module_index]))

        expected_rare_gene_modules = expected["rare_gene_modules"]
        assert actual_rare_gene_modules == expected_rare_gene_modules


def test_direct_pipeline() -> None:
    for path in glob("../metacells-test-data/*.h5ad"):
        adata, expected = _load(path)

        mc.ut.log_calc("path", path)
        pdata = adata[range(6000), :].copy()

        mc.pl.exclude_genes(pdata, random_seed=123456, **expected.get("exclude_genes", {}))
        mc.pl.exclude_cells(pdata, **expected.get("exclude_cells", {}))
        cdata = mc.pl.extract_clean_data(pdata)
        assert cdata is not None

        mc.pl.mark_lateral_genes(pdata, **expected.get("mark_lateral_genes", {}))

        mc.pl.compute_direct_metacells(cdata, random_seed=123456, **expected.get("compute_direct_metacells", {}))

        mdata = mc.pl.collect_metacells(cdata, random_seed=123456)

        mc.pl.compute_for_mcview(adata=cdata, gdata=mdata, random_seed=123456)

        expected_results = expected["inner_stdev_log"]

        actual_results = np.mean(
            mc.ut.to_numpy_matrix(mc.ut.get_vo_proper(mdata, "inner_stdev_log", layout="column_major"))
        )

        # mc.ut.log_calc('PATH', path)
        # mc.ut.log_calc('EXPECT', expected_results)
        # mc.ut.log_calc('ACTUAL', actual_results)
        assert np.allclose([expected_results], [actual_results])



"""
Documentation
-------------

Utilities for documenting Python functions.
"""

from inspect import Parameter
from inspect import signature
from typing import Any
from typing import Callable
from typing import TypeVar
from warnings import warn

__all__ = [
    "expand_doc",
]


CALLABLE = TypeVar("CALLABLE")


def expand_doc(**kwargs: Any) -> Callable[[CALLABLE], CALLABLE]:
    '''
    Expand the keyword arguments and the annotated function's default argument values inside the
    function's document string.

    That is, given something like:

    .. code:: python

        @expand_doc(foo=7)
        def bar(baz, vaz=5):
            """
            Bar with {foo} foos and parameter vaz (default: {baz}).
            """

    Then ``help(bar)`` will print:

    .. code:: text

        Bar with 7 foos and parameter vaz (default: 5).
    '''

    def documented(function: Callable) -> Callable:
        for parameter in signature(function).parameters.values():
            if parameter.default != Parameter.empty and parameter.name not in kwargs:
                kwargs[parameter.name] = parameter.default

        assert function.__doc__ is not None
        try:
            expanded_doc = function.__doc__.format_map(kwargs)
        except BaseException as exception:
            raise RuntimeError(  # pylint: disable=raise-missing-from
                f"key {exception} "
                f"in expand_doc documentation for the function "
                f"{function.__module__}.{function.__qualname__}"
            )

        if expanded_doc == function.__doc__:
            expand_doc_had_no_effect = (
                "@expand_doc had no effect on the documentation of the function "
                f"{function.__module__}.{function.__qualname__}"
            )
            warn(expand_doc_had_no_effect)

        function.__doc__ = expanded_doc
        return function

    return documented  # type: ignore



"""
Exclude
-------

Raw single-cell RNA sequencing data is notoriously noisy and "dirty". The pipeline steps here performs initial analysis
of the data and exclude some of it, so it would not harm the metacells computation. The steps provided here are expected
to be generically useful, but as always specific data sets may require custom cleaning steps on a case-by-case basis.
"""

from re import Pattern
from typing import Collection
from typing import List
from typing import Optional
from typing import Union

from anndata import AnnData  # type: ignore

import metacells.parameters as pr
import metacells.tools as tl
import metacells.utilities as ut

__all__ = [
    "exclude_genes",
    "exclude_cells",
    "extract_clean_data",
]


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def exclude_genes(
    adata: AnnData,
    what: Union[str, ut.Matrix] = "__x__",
    *,
    bursty_max_sampled_cells: Optional[int] = pr.bursty_lonely_max_sampled_cells,
    bursty_downsample_min_samples: int = pr.bursty_lonely_downsample_min_samples,
    bursty_downsample_min_cell_quantile: float = pr.bursty_lonely_downsample_max_cell_quantile,
    bursty_downsample_max_cell_quantile: float = pr.bursty_lonely_downsample_min_cell_quantile,
    bursty_min_gene_total: int = pr.bursty_lonely_min_gene_total,
    bursty_min_gene_normalized_variance: float = pr.bursty_lonely_min_gene_normalized_variance,
    bursty_max_gene_similarity: float = pr.bursty_lonely_max_gene_similarity,
    properly_sampled_min_gene_total: Optional[int] = pr.properly_sampled_min_gene_total,
    excluded_gene_names: Optional[Collection[str]] = None,
    excluded_gene_patterns: Optional[Collection[Union[str, Pattern]]] = None,
    random_seed: int,
) -> None:
    """
    Exclude a subset of the genes from the metacells computation.

    You can also just manually set the ``excluded_gene`` mask, or further manipulate it after calling this function.

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes.

    **Returns**

    Sets the following in the data:

    Variable (gene) annotations:
        ``properly_sampled_gene``
            A mask of the "properly sampled" genes.

        ``excluded_gene``
            A mask of the genes which were excluded (by name or due to not being properly sampled).

    **Computation Parameters**

    1. Invoke :py:func:`metacells.tools.bursty_lonely.find_bursty_lonely_genes` using
       ``bursty_max_sampled_cells`` (default: {bursty_max_sampled_cells}),
       ``bursty_downsample_min_samples`` (default: {bursty_downsample_min_samples}),
       ``bursty_downsample_min_cell_quantile`` (default: {bursty_downsample_min_cell_quantile}),
       ``bursty_downsample_max_cell_quantile`` (default: {bursty_downsample_max_cell_quantile}),
       ``bursty_min_gene_total`` (default: {bursty_min_gene_total}),
       ``bursty_min_gene_normalized_variance`` (default: {bursty_min_gene_normalized_variance}),
       ``bursty_max_gene_similarity`` (default: {bursty_max_gene_similarity}),
       and ``random_seed``.
       Any "bursty_lonely_genes" will be excluded.

    2. Invoke :py:func:`metacells.tools.properly_sampled.find_properly_sampled_genes` using
       ``properly_sampled_min_gene_total`` (default: {properly_sampled_min_gene_total}). Genes which are not properly
       sampled will be excluded.

    3. Invoke :py:func:`metacells.tools.named.find_named_genes` to also exclude genes based on their name, using the
       ``excluded_gene_names`` (default: {excluded_gene_names}) and ``excluded_gene_patterns`` (default:
       {excluded_gene_patterns}).
    """
    if (
        bursty_max_sampled_cells is None
        and properly_sampled_min_gene_total is None
        and excluded_gene_names is None
        and excluded_gene_patterns is None
    ):
        return

    bursty_lonely_genes_mask: Optional[ut.NumpyVector] = None
    if bursty_max_sampled_cells is not None:
        tl.find_bursty_lonely_genes(
            adata,
            what,
            max_sampled_cells=bursty_max_sampled_cells,
            downsample_min_samples=bursty_downsample_min_samples,
            downsample_min_cell_quantile=bursty_downsample_min_cell_quantile,
            downsample_max_cell_quantile=bursty_downsample_max_cell_quantile,
            min_gene_total=bursty_min_gene_total,
            min_gene_normalized_variance=bursty_min_gene_normalized_variance,
            max_gene_similarity=bursty_max_gene_similarity,
            random_seed=random_seed,
        )
        bursty_lonely_genes_mask = ut.get_v_numpy(adata, "bursty_lonely_gene")

    properly_sampled_genes_mask: Optional[ut.NumpyVector] = None
    if properly_sampled_min_gene_total is not None:
        tl.find_properly_sampled_genes(adata, what, min_gene_total=properly_sampled_min_gene_total)
        properly_sampled_genes_mask = ut.get_v_numpy(adata, "properly_sampled_gene")

    named_genes_mask: Optional[ut.NumpyVector] = None
    if excluded_gene_names is not None or excluded_gene_patterns is not None:
        named_genes_series = tl.find_named_genes(
            adata, names=excluded_gene_names, patterns=excluded_gene_patterns, to=None
        )
        assert named_genes_series is not None
        named_genes_mask = named_genes_series.values

    excluded_genes_mask: Optional[ut.NumpyVector] = None
    if bursty_lonely_genes_mask is not None:
        excluded_genes_mask = bursty_lonely_genes_mask

    if properly_sampled_genes_mask is not None:
        if excluded_genes_mask is None:
            excluded_genes_mask = ~properly_sampled_genes_mask
        else:
            excluded_genes_mask = excluded_genes_mask | ~properly_sampled_genes_mask

    if named_genes_mask is not None:
        if excluded_genes_mask is None:
            excluded_genes_mask = named_genes_mask
        else:
            excluded_genes_mask = excluded_genes_mask | named_genes_mask

    assert excluded_genes_mask is not None
    ut.set_v_data(adata, "excluded_gene", excluded_genes_mask)


@ut.logged()
@ut.timed_call()
def exclude_cells(
    adata: AnnData,
    what: Union[str, ut.Matrix] = "__x__",
    *,
    properly_sampled_min_cell_total: Optional[int],
    properly_sampled_max_cell_total: Optional[int],
    properly_sampled_max_excluded_genes_fraction: Optional[float],
    additional_cells_masks: Optional[List[str]] = None,
) -> None:
    """
    Exclude a subset of the cells from the metacells computation.

    You can also just manually set the ``excluded_cell`` mask, or further manipulate it after calling this function.

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes. Optionally, may contain an
    ``excluded_gene`` mask of genes to be excluded from the metacells computation. That is, invoke this after
    calling :py:func:`exclude_genes` (if you wish to exclude any genes).

    **Returns**

    Sets the following in the full data:

    Observation (cell) annotations:
        ``properly_sampled_cell``
            A mask of the "properly sampled" cells.

        ``excluded_cell``
            A mask of the genes which were excluded (inverse of ``properly_sampled_cell``).

    **Computation Parameters**

    1. Invoke :py:func:`metacells.tools.properly_sampled.find_properly_sampled_cells` using
       ``properly_sampled_min_cell_total`` (no default), ``properly_sampled_max_cell_total`` (no default) and
       ``properly_sampled_max_excluded_genes_fraction`` (no default).

    2. Exclude any cells which are not properly sampled (``|~properly_sampled_cell``), with optional additional
       following ``additional_cells_masks`` (using  :py:func:`metacells.tools.mask.combine_masks`).
    """
    tl.find_properly_sampled_cells(
        adata,
        what,
        min_cell_total=properly_sampled_min_cell_total,
        max_cell_total=properly_sampled_max_cell_total,
        max_excluded_genes_fraction=properly_sampled_max_excluded_genes_fraction,
    )

    excluded_cells_masks = ["|~properly_sampled_cell"] + (additional_cells_masks or [])
    tl.combine_masks(adata, excluded_cells_masks, to="excluded_cell")


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def extract_clean_data(
    adata: AnnData,
    *,
    name: Optional[str] = ".clean",
    top_level: bool = True,
) -> Optional[AnnData]:
    """
    Extract a "clean" subset of the ``adata`` to compute metacells for.

    .. todo::

        Allow computing metacells on the full data, obeying the ``excluded_cell`` and ``excluded_gene`` masks.

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes. This may contain per-gene
    ``excluded_gene`` and per-cell ``excluded_cell`` annotations.

    **Returns**

    Annotated sliced data containing the "clean" subset of the original data (that is, everything not
    marked as excluded).

    The returned data will have ``full_cell_index`` and ``full_gene_index`` per-observation (cell) and per-variable
    (gene) annotations to allow mapping the results back to the original data.

    **Computation Parameters**

    1. This simply :py:func:`metacells.tools.filter.filter_data` to slice just the genes not in the ``excluded_gene``
       mask and the cells not in the ``excluded_cell`` mask, data using the ``name`` (default: {name}), and tracking the
       original ``full_cell_index`` and ``full_gene_index``.
    """
    results = tl.filter_data(
        adata,
        name=name,
        top_level=top_level,
        track_obs="full_cell_index",
        track_var="full_gene_index",
        obs_masks=["&~excluded_cell?"],
        var_masks=["&~excluded_gene?"],
    )

    if results is None:
        return None
    return results[0]



"""
K-Nearest-Neighbors Graph
-------------------------
"""

from typing import Optional
from typing import Union

import numpy as np
import scipy.sparse as sp  # type: ignore
from anndata import AnnData  # type: ignore

import metacells.parameters as pr
import metacells.utilities as ut

__all__ = [
    "compute_obs_obs_knn_graph",
    "compute_var_var_knn_graph",
]


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def compute_obs_obs_knn_graph(
    adata: AnnData,
    what: Union[str, ut.Matrix] = "obs_similarity",
    *,
    k: int,
    balanced_ranks_factor: float = pr.knn_balanced_ranks_factor,
    incoming_degree_factor: float = pr.knn_incoming_degree_factor,
    outgoing_degree_factor: float = pr.knn_outgoing_degree_factor,
    min_outgoing_degree: int = pr.knn_min_outgoing_degree,
    inplace: bool = True,
) -> Optional[ut.PandasFrame]:
    """
    Compute a directed  K-Nearest-Neighbors graph based on ``what`` (default: what) similarity data
    for each pair of observations (cells).

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes, where
    ``what`` is a per-observation-per-observation matrix or the name of a
    per-observation-per-observation annotation containing such a matrix.

    **Returns**

    Observations-Pair Annotations
        ``obs_outgoing_weights``
            A sparse square matrix where each non-zero entry is the weight of an edge between a pair
            of cells or genes, where the sum of the weights of the outgoing edges for each element
            is 1 (there is always at least one such edge).

    If ``inplace`` (default: {inplace}), this is written to the data, and the function returns
    ``None``. Otherwise this is returned as a pandas data frame (indexed by the observation names).

    **Computation Parameters**

    1. Use the ``obs_similarity`` and convert it to ranks (in descending order).
       This gives us a dense asymmetric ``<elements>_outgoing_ranks`` matrix.

    2. Convert the asymmetric outgoing ranks matrix into a symmetric ``obs_balanced_ranks`` matrix
       by element-wise multiplying it with its transpose and taking the square root. That is, for
       each edge to be high-balanced-rank, the geomean of its outgoing rank has to be high in both
       nodes it connects.

       .. note::

            This can drastically reduce the degree of the nodes, since to survive an edge needs to
            have been in the top ranks for both its nodes (as multiplying with zero drops the edge).
            This is why the ``balanced_ranks_factor`` needs to be large-ish.

    3. Keeping only balanced ranks of geomean of up to ``k * balanced_ranks_factor`` (default:
       {balanced_ranks_factor}). This does a preliminary pruning of low-quality edges.

    4. Prune the edges, keeping only the ``k * incoming_degree_factor`` (default: k * {incoming_degree_factor})
       highest-ranked incoming edges for each node, and then only the ``k * outgoing_degree_factor`` (default:
       {outgoing_degree_factor}) highest-ranked outgoing edges for each node, while ensuring that the
       highest-balanced-ranked outgoing edge of each node is preserved. This gives us an asymmetric ``obs_pruned_ranks``
       matrix, which has the structure we want, but not the correct edge weights yet.

       .. note::

            Balancing the ranks, and then pruning the incoming edges, ensures that "hub" nodes, that
            is nodes that many other nodes prefer to connect with, end up connected to a limited
            number of such "spoke" nodes.

    5. If there is any node which is left with an out degree of less than ``min_outgoing_degree`` (default:
       {min_outgoing_degree}), increase K by 10% and repeat steps 2-4.

    6. Normalize the outgoing edge weights by dividing them with the sum of their balanced ranks,
       such that the sum of the outgoing edge weights for each node is 1. Note that there is always
       at least one outgoing edge for each node. This gives us the ``obs_outgoing_weights`` for our
       directed K-Nearest-Neighbors graph.

       .. note::

            Ensuring each node has at least one outgoing edge allows us to always have at least one
            candidate grouping to add it to. This of course doesn't protect the node from being
            rejected by its group as deviant.
    """
    return _compute_elements_knn_graph(
        adata,
        "obs",
        what,
        k=k,
        balanced_ranks_factor=balanced_ranks_factor,
        incoming_degree_factor=incoming_degree_factor,
        outgoing_degree_factor=outgoing_degree_factor,
        min_outgoing_degree=min_outgoing_degree,
        inplace=inplace,
    )


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def compute_var_var_knn_graph(
    adata: AnnData,
    what: Union[str, ut.Matrix] = "var_similarity",
    *,
    k: int,
    balanced_ranks_factor: float = pr.knn_balanced_ranks_factor,
    incoming_degree_factor: float = pr.knn_incoming_degree_factor,
    outgoing_degree_factor: float = pr.knn_outgoing_degree_factor,
    min_outgoing_degree: int = pr.knn_min_outgoing_degree,
    inplace: bool = True,
) -> Optional[ut.PandasFrame]:
    """
    Compute a directed  K-Nearest-Neighbors graph based on ``what`` (default: what) similarity data
    for each pair of variables (genes).

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes, where
    ``what`` is a per-variable-per-variable matrix or the name of a per-variable-per-variable
    annotation containing such a matrix.

    **Returns**

    Variables-Pair Annotations
        ``var_outgoing_weights``
            A sparse square matrix where each non-zero entry is the weight of an edge between a pair
            of cells or genes, where the sum of the weights of the outgoing edges for each element
            is 1 (there is always at least one such edge).

    If ``inplace`` (default: {inplace}), this is written to the data, and the function returns
    ``None``. Otherwise this is returned as a pandas data frame (indexed by the variable names).

    **Computation Parameters**

    1. Use the ``var_similarity`` and convert it to ranks (in descending order).
       This gives us a dense asymmetric ``<elements>_outgoing_ranks`` matrix.

    2. Convert the asymmetric outgoing ranks matrix into a symmetric ``var_balanced_ranks`` matrix
       by element-wise multiplying it with its transpose and taking the square root. That is, for
       each edge to be high-balanced-rank, the geomean of its outgoing rank has to be high in both
       nodes it connects.

    3. Keeping only balanced ranks of up to ``k * k * balanced_ranks_factor`` (default:
       {balanced_ranks_factor}). This does a preliminary pruning of low-quality edges.

    4. Prune the edges, keeping only the ``k * incoming_degree_factor`` (default: k * {incoming_degree_factor})
       highest-ranked incoming edges for each node, and then only the ``k * outgoing_degree_factor`` (default:
       {outgoing_degree_factor}) highest-ranked outgoing edges for each node, while ensuring that the
       highest-balanced-ranked outgoing edge of each node is preserved. This gives us an asymmetric ``var_pruned_ranks``
       matrix, which has the structure we want, but not the correct edge weights yet.

       .. note::

            Balancing the ranks, and then pruning the incoming edges, ensures that "hub" nodes, that
            is nodes that many other nodes prefer to connect with, end up connected to a limited
            number of such "spoke" nodes.

    5. If there is any node which is left with an out degree of less than ``min_outgoing_degree`` (default:
       {min_outgoing_degree}), increase K by 10% and repeat steps 2-4.

    6. Normalize the outgoing edge weights by dividing them with the sum of their balanced ranks,
       such that the sum of the outgoing edge weights for each node is 1. Note that there is always
       at least one outgoing edge for each node. This gives us the ``var_outgoing_weights`` for our
       directed K-Nearest-Neighbors graph.

       .. note::

            Ensuring each node has at least one outgoing edge allows us to always have at least one
            candidate grouping to add it to. This of course doesn't protect the node from being
            rejected by its group as deviant.
    """
    return _compute_elements_knn_graph(
        adata,
        "var",
        what,
        k=k,
        balanced_ranks_factor=balanced_ranks_factor,
        incoming_degree_factor=incoming_degree_factor,
        outgoing_degree_factor=outgoing_degree_factor,
        min_outgoing_degree=min_outgoing_degree,
        inplace=inplace,
    )


def _compute_elements_knn_graph(
    adata: AnnData,
    elements: str,
    what: Union[str, ut.Matrix] = "__x__",
    *,
    k: int,
    balanced_ranks_factor: float,
    incoming_degree_factor: float,
    outgoing_degree_factor: float,
    min_outgoing_degree: int,
    inplace: bool = True,
) -> Optional[ut.PandasFrame]:
    assert elements in ("obs", "var")
    assert balanced_ranks_factor > 0.0
    assert incoming_degree_factor > 0.0
    assert outgoing_degree_factor > 0.0
    assert min_outgoing_degree >= 1

    if elements == "obs":
        get_data = ut.get_oo_proper
        set_data = ut.set_oo_data
    else:
        get_data = ut.get_vv_proper
        set_data = ut.set_vv_data

    def store_matrix(matrix: ut.CompressedMatrix, name: str, when: bool) -> None:  #
        if when:
            name = elements + "_" + name
            set_data(
                adata,
                name,
                matrix,
                formatter=lambda matrix: ut.ratio_description(
                    matrix.shape[0] * matrix.shape[1], "element", matrix.nnz, "nonzero"
                ),
            )
        elif ut.logging_calc():
            ut.log_calc(
                f"{elements}_{name}",
                ut.ratio_description(matrix.shape[0] * matrix.shape[1], "element", matrix.nnz, "nonzero"),
            )

    similarity = ut.to_proper_matrix(get_data(adata, what))
    similarity = ut.to_layout(similarity, "row_major", symmetric=True)
    similarity = ut.to_numpy_matrix(similarity)

    ut.log_calc("similarity", similarity)

    while True:
        outgoing_ranks = _rank_outgoing(similarity)

        balanced_ranks = _balance_ranks(outgoing_ranks, k, balanced_ranks_factor)
        store_matrix(balanced_ranks, "balanced_ranks", True)

        pruned_ranks = _prune_ranks(balanced_ranks, k, incoming_degree_factor, outgoing_degree_factor)
        store_matrix(pruned_ranks, "pruned_ranks", True)

        outgoing_ranks = ut.nnz_per(pruned_ranks, per="row")
        outgoing_degree = np.min(outgoing_ranks)
        ut.log_calc("outgoing_degree", outgoing_degree)

        if outgoing_degree >= min_outgoing_degree:
            break

        k += max(1, int(k / 10))
        ut.log_calc("k", k)

    outgoing_weights = _weigh_edges(pruned_ranks)
    store_matrix(outgoing_weights, "outgoing_weights", inplace)

    if inplace:
        return None

    if elements == "obs":
        names = adata.obs_names
    else:
        names = adata.var_names

    return ut.to_pandas_frame(outgoing_weights, index=names, columns=names)


@ut.timed_call()
def _rank_outgoing(similarity: ut.NumpyMatrix) -> ut.NumpyMatrix:
    size = similarity.shape[0]
    assert similarity.shape == (size, size)
    similarity = np.copy(similarity)

    min_similarity = ut.min_matrix(similarity)

    np.fill_diagonal(similarity, min_similarity - 1)

    assert ut.is_layout(similarity, "row_major")
    outgoing_ranks = ut.rank_matrix_by_layout(similarity, ascending=False)
    assert np.sum(np.diagonal(outgoing_ranks) == size) == size
    return outgoing_ranks


@ut.timed_call()
def _balance_ranks(outgoing_ranks: ut.NumpyMatrix, k: int, balanced_ranks_factor: float) -> ut.CompressedMatrix:
    size = outgoing_ranks.shape[0]

    with ut.timed_step(".multiply"):
        ut.timed_parameters(size=size)
        dense_balanced_ranks = outgoing_ranks
        assert np.sum(np.diagonal(dense_balanced_ranks) == size) == size
        dense_balanced_ranks *= outgoing_ranks.transpose()

    with ut.timed_step(".sqrt"):
        np.sqrt(dense_balanced_ranks, out=dense_balanced_ranks)

    max_rank = k * balanced_ranks_factor
    ut.log_calc("max_rank", max_rank)

    dense_balanced_ranks *= -1
    dense_balanced_ranks += 2**21

    with ut.timed_step("numpy.argmax"):
        ut.timed_parameters(size=size)
        max_index_of_each = ut.to_numpy_vector(dense_balanced_ranks.argmax(axis=1))  #

    dense_balanced_ranks += max_rank + 1 - 2**21

    preserved_row_indices = np.arange(size)
    preserved_column_indices = max_index_of_each
    preserved_balanced_ranks = ut.to_numpy_vector(dense_balanced_ranks[preserved_row_indices, preserved_column_indices])

    preserved_balanced_ranks[preserved_balanced_ranks < 1] = 1

    dense_balanced_ranks[dense_balanced_ranks < 0] = 0
    np.fill_diagonal(dense_balanced_ranks, 0)

    dense_balanced_ranks[preserved_row_indices, preserved_column_indices] = preserved_balanced_ranks

    assert np.sum(np.diagonal(dense_balanced_ranks) == 0) == size
    sparse_balanced_ranks = sp.csr_matrix(dense_balanced_ranks)

    _assert_proper_compressed(sparse_balanced_ranks, "csr")
    return sparse_balanced_ranks


@ut.timed_call()
def _prune_ranks(
    balanced_ranks: ut.CompressedMatrix, k: int, incoming_degree_factor: float, outgoing_degree_factor: float
) -> ut.CompressedMatrix:
    size = balanced_ranks.shape[0]

    incoming_degree = int(round(k * incoming_degree_factor))
    incoming_degree = min(incoming_degree, size - 1)
    ut.log_calc("incoming_degree", incoming_degree)

    outgoing_degree = int(round(k * outgoing_degree_factor))
    outgoing_degree = min(outgoing_degree, size - 1)
    ut.log_calc("outgoing_degree", outgoing_degree)

    all_indices = np.arange(size)
    with ut.timed_step("numpy.argmax"):
        ut.timed_parameters(results=size, elements=balanced_ranks.nnz / size)
        max_index_of_each = ut.to_numpy_vector(balanced_ranks.argmax(axis=1))

    preserved_row_indices = all_indices
    preserved_column_indices = max_index_of_each
    preserved_balanced_ranks = ut.to_numpy_vector(balanced_ranks[preserved_row_indices, preserved_column_indices])
    assert np.min(preserved_balanced_ranks) > 0
    preserved_matrix = sp.coo_matrix(
        (preserved_balanced_ranks, (preserved_row_indices, preserved_column_indices)), shape=balanced_ranks.shape
    )
    preserved_matrix.has_canonical_format = True

    pruned_ranks = ut.mustbe_compressed_matrix(ut.to_layout(balanced_ranks, "column_major", symmetric=True))
    _assert_proper_compressed(pruned_ranks, "csc")

    pruned_ranks = ut.prune_per(pruned_ranks, incoming_degree)
    _assert_proper_compressed(pruned_ranks, "csc")

    pruned_ranks = ut.mustbe_compressed_matrix(ut.to_layout(pruned_ranks, "row_major"))
    _assert_proper_compressed(pruned_ranks, "csr")

    pruned_ranks = ut.prune_per(pruned_ranks, outgoing_degree)
    _assert_proper_compressed(pruned_ranks, "csr")

    with ut.timed_step("sparse.maximum"):
        ut.timed_parameters(collected=pruned_ranks.nnz, preserved=preserved_matrix.nnz)
        pruned_ranks = pruned_ranks.maximum(preserved_matrix)
        pruned_ranks = pruned_ranks.maximum(preserved_matrix.transpose())

    ut.sort_compressed_indices(pruned_ranks)

    pruned_ranks = ut.mustbe_compressed_matrix(pruned_ranks)
    _assert_proper_compressed(pruned_ranks, "csr")
    return pruned_ranks


@ut.timed_call()
def _weigh_edges(pruned_ranks: ut.CompressedMatrix) -> ut.CompressedMatrix:
    size = pruned_ranks.shape[0]

    total_ranks_per_row = ut.sum_per(pruned_ranks, per="row")

    ut.timed_parameters(size=size)
    scale_per_row = np.reciprocal(total_ranks_per_row, out=total_ranks_per_row)
    edge_weights = pruned_ranks.multiply(scale_per_row[:, None])
    edge_weights = ut.to_layout(edge_weights, "row_major")

    _assert_proper_compressed(edge_weights, "csr")
    return edge_weights


def _assert_proper_compressed(matrix: ut.CompressedMatrix, layout: str) -> None:
    assert sp.issparse(matrix)
    assert ut.shaped_dtype(matrix) == "float32"
    assert matrix.getformat() == layout
    assert matrix.has_sorted_indices
    assert matrix.has_canonical_format



"""
Apply
-----
"""

from typing import Any
from typing import Callable
from typing import Dict
from typing import NamedTuple
from typing import Optional
from typing import Union

import numpy as np
from anndata import AnnData  # type: ignore

import metacells.utilities as ut

__all__ = [
    "DefaultValues",
    "Skip",
    "Raise",
    "apply_obs_annotations",
    "apply_var_annotations",
]


class DefaultValues(NamedTuple):
    """
    Default values to use in :py:func:`apply_obs_annotations` and :py:func:`apply_var_annotations`.
    """

    #: The default value to use for the slice data.
    slice: Any

    #: The default value to use for the full data.
    full: Any

    def __str__(self) -> str:
        return f"slice: {self.slice} full: {self.full}"


class Skip:
    """
    A special value indicating to skip the annotation if it does not exist.
    """


class Raise:
    """
    A special value indicating to raise a ``KeyError`` if an annotation does not exist.
    """


@ut.logged(annotations=str)
@ut.timed_call()
def apply_obs_annotations(
    adata: AnnData,
    sdata: AnnData,
    annotations: Dict[str, DefaultValues],
    *,
    indices: Union[str, ut.Vector],
) -> None:
    """
    Apply per-observation (cell) annotations of a slice ``sdata`` to the full ``adata``.

    **Input**

    Annotated ``adata``, and a slice of it ``sdata``, where the ``indices`` is either the vector of
    full indices of the slice observations, or the name of a per-observation annotation of ``sdata``
    that contains this vector.

    **Computation Parameters**

    1. Loop on each of the named ``annotations``, where the value associated with the
       name is used as the default value (see below).

    2. If the slice data does not contain a per-observation (cell) annotation of this name, consider
       the :py:attr:`DefaultValues.slice`:

       * If it is :py:const:`Raise`, raise a ``KeyError``.

       * If it is :py:const:`Skip`, do not apply the annotation to the full data.

       * Otherwise, behave as if the annotation's value was a vector containing the
         :py:attr:`DefaultValues.slice` value.

    3. If the full data does not contain a per-observation (cell) annotation of this name, consider
       the :py:attr:`DefaultValues.full`:

       * If it is :py:const:`Raise`, raise a ``KeyError``.

       * If it is :py:const:`Skip`, do not apply the annotation to the full data.

       * Otherwise, initialize the annotation to a vector containing the
         :py:attr:`DefaultValues.full` value.

    4. Apply the slice data values to the entries of the full data identified by the ``indices``.
    """
    _apply_annotations(adata, sdata, "o", annotations, indices)


@ut.logged(annotations=str)
@ut.timed_call()
def apply_var_annotations(
    adata: AnnData,
    sdata: AnnData,
    annotations: Dict[str, DefaultValues],
    *,
    indices: Union[str, ut.Vector],
) -> None:
    """
    Apply per-variable (gene) annotations of a slice ``sdata`` to the full ``adata``.

    **Input**

    Annotated ``adata``, and a slice of it ``sdata``, where the ``indices`` is either the vector of
    full indices of the slice variables, or the name of a per-variable annotation of ``sdata`` that
    contains this vector.

    **Computation Parameters**

    1. Loop on each of the named ``annotations``, where the value associated with the
       name is used as the default value (see below).

    2. If the slice data does not contain a per-variable (gene) annotation of this name, consider
       the :py:attr:`DefaultValues.slice`:

       * If it is :py:const:`Raise`, raise a ``KeyError``.

       * If it is :py:const:`Skip`, do not apply the annotation to the full data.

       * Otherwise, behave as if the annotation's value was a vector containing the
         :py:attr:`DefaultValues.slice` value.

    3. If the full data does not contain a per-variable (gene) variable of this name, consider
       the :py:attr:`DefaultValues.full`:

       * If it is :py:const:`Raise`, raise a ``KeyError``.

       * If it is :py:const:`Skip`, do not apply the annotation to the full data.

       * Otherwise, initialize the annotation to a vector containing the
         :py:attr:`DefaultValues.full` value.

    4. Apply the slice data values to the entries of the full data identified by the ``indices``.
    """
    _apply_annotations(adata, sdata, "v", annotations, indices)


def _apply_annotations(  # pylint: disable=too-many-branches
    adata: AnnData,
    sdata: AnnData,
    per: str,
    annotations: Dict[str, DefaultValues],
    indices: Union[str, ut.Vector],
) -> None:
    full_name = ut.get_name(adata)
    slice_name = ut.get_name(sdata)

    assert per in ("o", "v")

    if per == "o":
        full_data = adata.obs
        full_size = adata.n_obs
        slice_data = sdata.obs
        slice_size = sdata.n_obs
        full_indices = ut.get_o_numpy(sdata, indices)
    else:
        full_data = adata.var
        full_size = adata.n_vars
        slice_data = sdata.var
        slice_size = sdata.n_vars
        full_indices = ut.get_v_numpy(sdata, indices)

    for name, default_values in annotations.items():
        slice_value = slice_data.get(name)
        if slice_value is not None:
            formatter: Optional[Callable[[Any], str]] = None
        else:
            if default_values.slice == Skip or isinstance(default_values.slice, Skip):
                continue

            if default_values.slice == Raise or isinstance(default_values.slice, Raise):
                if slice_name is None:
                    raise KeyError(f"unknown slice data name: {name}")
                raise KeyError(f"unknown slice data: {slice_name} name: {name}")

            slice_value = default_values.slice

            def formatter(_: Any) -> str:
                # pylint: disable=cell-var-from-loop
                return f"{slice_size} <- {slice_value}"

            # pylint: enable=cell-var-from-loop

        full_value = full_data.get(name)
        if full_value is not None:
            ut.unfreeze(full_value)
        else:
            if default_values.full == Skip or isinstance(default_values.full, Skip):
                continue

            if default_values.full == Raise or isinstance(default_values.full, Raise):
                if full_name is None:
                    raise KeyError(f"unknown full data name: {name}")
                raise KeyError(f"unknown full data: {full_name} name: {name}")

            if default_values.full is None:
                full_value = np.full(full_size, None, dtype="float32")
            else:
                full_value = np.full(full_size, default_values.full)

        full_value[full_indices] = slice_value
        if per == "o":
            ut.set_o_data(adata, name, full_value, formatter=formatter)
        else:
            ut.set_v_data(adata, name, full_value, formatter=formatter)



"""
Direct
------
"""

from typing import Optional
from typing import Union

import numpy as np
from anndata import AnnData  # type: ignore

import metacells.parameters as pr
import metacells.tools as tl
import metacells.utilities as ut

from .select import extract_selected_data

__all__ = [
    "compute_direct_metacells",
]


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def compute_direct_metacells(
    adata: AnnData,
    what: Union[str, ut.Matrix] = "__x__",
    *,
    select_downsample_min_samples: int = pr.select_downsample_min_samples,
    select_downsample_min_cell_quantile: float = pr.select_downsample_min_cell_quantile,
    select_downsample_max_cell_quantile: float = pr.select_downsample_max_cell_quantile,
    select_min_gene_total: Optional[int] = pr.select_min_gene_total,
    select_min_gene_top3: Optional[int] = pr.select_min_gene_top3,
    select_min_gene_relative_variance: Optional[float] = pr.select_min_gene_relative_variance,
    select_min_genes: int = pr.select_min_genes,
    cells_similarity_value_regularization: float = pr.cells_similarity_value_regularization,
    cells_similarity_log_data: bool = pr.cells_similarity_log_data,
    cells_similarity_method: str = pr.cells_similarity_method,
    target_metacell_size: int = pr.target_metacell_size,
    min_metacell_size: int = pr.min_metacell_size,
    target_metacell_umis: int = pr.target_metacell_umis,
    cell_umis: Optional[ut.NumpyVector] = pr.cell_umis,
    knn_k: Optional[int] = pr.knn_k,
    knn_k_size_factor: float = pr.candidates_knn_k_size_factor,
    knn_k_umis_quantile: float = pr.knn_k_umis_quantile,
    min_knn_k: Optional[int] = pr.min_knn_k,
    knn_balanced_ranks_factor: float = pr.knn_balanced_ranks_factor,
    knn_incoming_degree_factor: float = pr.knn_incoming_degree_factor,
    knn_outgoing_degree_factor: float = pr.knn_outgoing_degree_factor,
    knn_min_outgoing_degree: int = pr.knn_min_outgoing_degree,
    min_seed_size_quantile: float = pr.min_seed_size_quantile,
    max_seed_size_quantile: float = pr.max_seed_size_quantile,
    candidates_cooldown_pass: float = pr.cooldown_pass,
    candidates_cooldown_node: float = pr.cooldown_node,
    candidates_cooldown_phase: float = pr.cooldown_phase,
    candidates_min_split_size_factor: float = pr.candidates_min_split_size_factor,
    candidates_max_merge_size_factor: float = pr.candidates_max_merge_size_factor,
    candidates_max_split_min_cut_strength: float = pr.max_split_min_cut_strength,
    candidates_min_cut_seed_cells: int = pr.min_cut_seed_cells,
    must_complete_cover: bool = False,
    deviants_policy: str = pr.deviants_policy,
    deviants_min_gene_fold_factor: float = pr.deviants_min_gene_fold_factor,
    deviants_gap_skip_cells: int = pr.deviants_gap_skip_cells,
    deviants_min_noisy_gene_fold_factor: float = pr.deviants_min_noisy_gene_fold_factor,
    deviants_max_gene_fraction: float = pr.deviants_max_gene_fraction,
    deviants_max_cell_fraction: Optional[float] = pr.deviants_max_cell_fraction,
    deviants_max_gap_cells_count: int = pr.deviants_max_gap_cells_count,
    deviants_max_gap_cells_fraction: float = pr.deviants_max_gap_cells_fraction,
    dissolve_min_robust_size_factor: float = pr.dissolve_min_robust_size_factor,
    dissolve_min_convincing_gene_fold_factor: Optional[float] = pr.dissolve_min_convincing_gene_fold_factor,
    random_seed: int,
) -> None:
    """
    Directly compute metacells using ``what`` (default: {what}) data.

    This directly computes the metacells on the whole data. Like any method that directly looks at
    the whole data at once, the amount of CPU and memory needed becomes unreasonable when the data
    size grows. Above O(10,000) you are much better off using the divide-and-conquer method.

    .. note::

        The current implementation is naive in that it computes the full dense N^2 correlation
        matrix, and only then extracts the sparse graph out of it. We actually need two copies where
        each requires 4 bytes per entry, so for O(100,000) cells, we have storage of
        O(100,000,000,000). In addition, the implementation is serial for the graph clustering
        phases.

        It is possible to mitigate this by fusing the correlations phase and the graph generation
        phase, parallelizing the result, and also (somehow) parallelizing the graph clustering
        phase. This might increase the "reasonable" size for the direct approach to O(100,000).

        We have decided not to invest in this direction since it won't allow us to push the size to
        O(1,000,000) and above. Instead we provide the divide-and-conquer method, which easily
        scales to O(1,000,000) on a single multi-core server, and to "unlimited" size if we further
        enhance the implementation to use a distributed compute cluster of such servers.

    .. todo::

        Should :py:func:`compute_direct_metacells` avoid computing the graph and partition it for a
        very small number of cells?

    **Input**

    The presumably "clean" annotated ``adata``, where the observations are cells and the variables
    are genes, where ``what`` is a per-variable-per-observation matrix or the name of a
    per-variable-per-observation annotation containing such a matrix.

    **Returns**

    Sets the following annotations in ``adata``:

    Unstructured Annotations
        ``downsample_samples``
            The target total number of samples in each downsampled cell.

    Observation-Variable (Cell-Gene) Annotations:
        ``downsampled``
            The downsampled data where the total number of samples in each cell is at most ``downsample_samples``.

    Variable (Gene) Annotations
        ``high_total_gene``
            A boolean mask of genes with "high" expression level (unless a ``select_gene`` mask exists).

        ``high_relative_variance_gene``
            A boolean mask of genes with "high" normalized variance, relative to other genes with a similar expression
            level (unless a ``select_gene`` mask exists).

        ``selected_gene``
            A boolean mask of the actually selected genes.

    Observation (Cell) Annotations
        ``metacell``
            The integer index of the metacell each cell belongs to. The metacells are in no
            particular order. Cells with no metacell assignment ("outliers") are given a metacell
            index of ``-1``.

    **Computation Parameters**

    0. If ``cell_umis`` is not specified, use the sum of the ``what`` data for each cell.

    1. Invoke :py:func:`metacells.pipeline.select.extract_selected_data` to extract "select" data
       from the clean data, using the
       ``select_downsample_min_samples`` (default: {select_downsample_min_samples}),
       ``select_downsample_min_cell_quantile`` (default: {select_downsample_min_cell_quantile}),
       ``select_downsample_max_cell_quantile`` (default: {select_downsample_max_cell_quantile}),
       ``select_min_gene_total`` (default: {select_min_gene_total}),
       ``select_min_gene_top3`` (default: {select_min_gene_top3}),
       ``select_min_gene_relative_variance`` (default: {select_min_gene_relative_variance}) and
       ``random_seed``.

    2. Compute the fractions of each variable in each cell, and add the
       ``cells_similarity_value_regularization`` (default: {cells_similarity_value_regularization}) to
       it.

    3. If ``cells_similarity_log_data`` (default: {cells_similarity_log_data}), invoke the
       :py:func:`metacells.utilities.computation.log_data` function to compute the log (base 2) of
       the data.

    4. Invoke :py:func:`metacells.tools.similarity.compute_obs_obs_similarity` to compute the
       similarity between each pair of cells, using the
       ``cells_similarity_method`` (default: {cells_similarity_method}).

    5. Invoke :py:func:`metacells.tools.knn_graph.compute_obs_obs_knn_graph` to compute a K-Nearest-Neighbors graph,
       using the ``knn_balanced_ranks_factor`` (default: {knn_balanced_ranks_factor}), ``knn_incoming_degree_factor``
       (default: {knn_incoming_degree_factor}) ``knn_outgoing_degree_factor`` (default: {knn_outgoing_degree_factor})
       and ``knn_min_outgoing_degree`` (default: {knn_min_outgoing_degree})(. If ``knn_k`` (default: {knn_k}) is not
       specified, then it is chosen to be the ``knn_k_size_factor`` (default: {knn_k_size_factor} times the median
       number of cells required to reach the target metacell size, or the ``knn_k_umis_quantile`` (default:
       {knn_k_umis_quantile}) the number of cells required, or ``min_knn_k`` (default: {min_knn_k}), whichever is
       largest.

    6. Invoke :py:func:`metacells.tools.candidates.compute_candidate_metacells` to compute
       the candidate metacells, using the
       ``min_seed_size_quantile`` (default: {min_seed_size_quantile}),
       ``max_seed_size_quantile`` (default: {max_seed_size_quantile}),
       ``candidates_cooldown_pass`` (default: {candidates_cooldown_pass}),
       ``candidates_cooldown_node`` (default: {candidates_cooldown_node}),
       ``candidates_cooldown_phase`` (default: {candidates_cooldown_phase}),
       ``candidates_min_split_size_factor`` (default: {candidates_min_split_size_factor}),
       ``candidates_max_merge_size_factor`` (default: {candidates_max_merge_size_factor}),
       and
       ``random_seed``. This tries to build metacells of the
       ``target_metacell_size`` (default: {target_metacell_size}),
       ``min_metacell_size`` (default: {min_metacell_size}), and
       ``target_metacell_umis`` (default: {target_metacell_umis})
       using the ``cell_umis`` (default: {cell_umis}).

    7. Unless ``must_complete_cover`` (default: {must_complete_cover}), invoke
       :py:func:`metacells.tools.deviants.find_deviant_cells` to remove deviants from the candidate
       metacells, using the
       ``deviants_policy`` (default: {deviants_policy}),
       ``deviants_min_gene_fold_factor`` (default: {deviants_min_gene_fold_factor}),
       ``deviants_gap_skip_cells`` (default: {deviants_gap_skip_cells}),
       ``deviants_min_noisy_gene_fold_factor`` (default: {deviants_min_noisy_gene_fold_factor}),
       ``deviants_max_gene_fraction`` (default: {deviants_max_gene_fraction})
       and
       ``deviants_max_cell_fraction`` (default: {deviants_max_cell_fraction}).

    9. Unless ``must_complete_cover`` (default: {must_complete_cover}), invoke
       :py:func:`metacells.tools.dissolve.dissolve_metacells` to dissolve small unconvincing
       metacells, using the same
       ``target_metacell_size`` (default: {target_metacell_size}),
       ``min_metacell_size`` (default: {min_metacell_size}),
       ``target_metacell_umis`` (default: {target_metacell_umis}),
       ``cell_umis`` (default: {cell_umis})
       and the
       ``dissolve_min_robust_size_factor`` (default: {dissolve_min_robust_size_factor}).
    """
    assert (
        target_metacell_size < 1000
    ), f"target_metacell_size: {target_metacell_size} seems to be in UMIs, should be in cells"

    if cell_umis is None:
        cell_umis = ut.get_o_numpy(adata, what, sum=True, formatter=ut.sizes_description).astype("float32")
    else:
        assert cell_umis.dtype == "float32"
    assert isinstance(cell_umis, ut.NumpyVector)

    if (
        adata.n_obs <= 2 * min(min_metacell_size, target_metacell_size * candidates_max_merge_size_factor)
        or np.sum(cell_umis) <= 2 * target_metacell_umis * candidates_max_merge_size_factor
    ):
        candidate_of_cells = np.zeros(adata.n_obs, dtype="int32")
        ut.set_v_data(adata, "selected_gene", np.zeros(adata.n_vars, dtype="bool"))

    else:
        sdata = extract_selected_data(
            adata,
            what,
            top_level=False,
            downsample_min_samples=select_downsample_min_samples,
            downsample_min_cell_quantile=select_downsample_min_cell_quantile,
            downsample_max_cell_quantile=select_downsample_max_cell_quantile,
            min_gene_relative_variance=select_min_gene_relative_variance,
            min_gene_total=select_min_gene_total,
            min_gene_top3=select_min_gene_top3,
            min_genes=select_min_genes,
            random_seed=random_seed,
        )

        data = ut.get_vo_proper(sdata, "downsampled", layout="row_major")
        data = ut.to_numpy_matrix(data, copy=True)

        if cells_similarity_value_regularization > 0:
            data += cells_similarity_value_regularization

        if cells_similarity_log_data:
            data = ut.log_data(data, base=2)

        if knn_k is None:
            knn_k_by_size = int(round(adata.n_obs / target_metacell_size))
            ut.log_calc("knn_k by size", knn_k_by_size)

            median_cell_umis = float(np.median(cell_umis))
            ut.log_calc("median_cell_umis", median_cell_umis)
            knn_k_of_median = int(round(target_metacell_umis / median_cell_umis))
            ut.log_calc("knn_k of median_cell_umis", knn_k_of_median)
            knn_k_by_median = int(round(knn_k_size_factor * target_metacell_umis / median_cell_umis))
            ut.log_calc("knn_k by median_cell_umis", knn_k_by_median)

            quantile_cell_umis = np.quantile(cell_umis, knn_k_umis_quantile)
            ut.log_calc("quantile_cell_umis", quantile_cell_umis)
            knn_k_by_quantile = int(round(target_metacell_umis / quantile_cell_umis))
            ut.log_calc("knn_k by quantile_cell_umis", knn_k_by_quantile)

        knn_k = max(knn_k_by_size, knn_k_by_median, knn_k_by_quantile, min_knn_k or 0)
        if knn_k == 0:
            ut.log_calc("knn_k: 0 (too small, trying a single metacell)")
            ut.set_o_data(sdata, "candidate", np.full(sdata.n_obs, 0, dtype="int32"), formatter=lambda _: "* <- 0")
        elif knn_k_of_median >= sdata.n_obs:
            ut.log_calc(f"knn_k of median: {knn_k_of_median} (too large, trying a single metacell)")
            ut.set_o_data(sdata, "candidate", np.full(sdata.n_obs, 0, dtype="int32"), formatter=lambda _: "* <- 0")

        else:
            ut.log_calc("knn_k", knn_k)

            tl.compute_obs_obs_similarity(sdata, data, method=cells_similarity_method, reproducible=random_seed != 0)

            tl.compute_obs_obs_knn_graph(
                sdata,
                k=knn_k,
                balanced_ranks_factor=knn_balanced_ranks_factor,
                incoming_degree_factor=knn_incoming_degree_factor,
                outgoing_degree_factor=knn_outgoing_degree_factor,
                min_outgoing_degree=knn_min_outgoing_degree,
            )

            tl.compute_candidate_metacells(
                sdata,
                target_metacell_size=target_metacell_size,
                min_metacell_size=min_metacell_size,
                target_metacell_umis=target_metacell_umis,
                cell_umis=cell_umis,
                min_seed_size_quantile=min_seed_size_quantile,
                max_seed_size_quantile=max_seed_size_quantile,
                cooldown_pass=candidates_cooldown_pass,
                cooldown_node=candidates_cooldown_node,
                cooldown_phase=candidates_cooldown_phase,
                min_split_size_factor=candidates_min_split_size_factor,
                max_merge_size_factor=candidates_max_merge_size_factor,
                max_split_min_cut_strength=candidates_max_split_min_cut_strength,
                min_cut_seed_cells=candidates_min_cut_seed_cells,
                must_complete_cover=must_complete_cover,
                random_seed=random_seed,
            )

        candidate_of_cells = ut.get_o_numpy(sdata, "candidate", formatter=ut.groups_description)

    if must_complete_cover:
        assert np.min(candidate_of_cells) == 0
        ut.set_o_data(adata, "metacell", candidate_of_cells, formatter=ut.groups_description)

    else:
        deviants = tl.find_deviant_cells(
            adata,
            candidates=candidate_of_cells,
            policy=deviants_policy,
            min_gene_fold_factor=deviants_min_gene_fold_factor,
            gap_skip_cells=deviants_gap_skip_cells,
            min_noisy_gene_fold_factor=deviants_min_noisy_gene_fold_factor,
            max_gene_fraction=deviants_max_gene_fraction,
            max_cell_fraction=deviants_max_cell_fraction,
            max_gap_cells_count=deviants_max_gap_cells_count,
            max_gap_cells_fraction=deviants_max_gap_cells_fraction,
        )

        tl.dissolve_metacells(
            adata,
            candidates=candidate_of_cells,
            deviants=deviants,
            target_metacell_size=target_metacell_size,
            min_metacell_size=min_metacell_size,
            target_metacell_umis=target_metacell_umis,
            cell_umis=cell_umis,
            min_robust_size_factor=dissolve_min_robust_size_factor,
            min_convincing_gene_fold_factor=dissolve_min_convincing_gene_fold_factor,
        )



"""
Metacells single-cell RNA sequencing.
"""

import sys

__author__ = "Oren Ben-Kiki"
__email__ = "oren@ben-kiki.org"
__version__ = "0.9.3"

if "sphinx" not in sys.argv[0]:
    from .should_check_avx2 import SHOULD_CHECK_AVX2

    if SHOULD_CHECK_AVX2:
        from . import check_avx2

# pylint: disable=wrong-import-position

from . import pipeline as pl
from . import tools as tl
from . import utilities as ut



"""
MCView
------

Compute metacell analysis in preparation for exporting the data to MCView.
"""

from typing import Any
from typing import Dict
from typing import Optional
from typing import Union

import numpy as np
from anndata import AnnData  # type: ignore

import metacells.tools as tl
import metacells.utilities as ut

from .umap import compute_umap_by_markers

__all__ = [
    "compute_for_mcview",
]


# pylint: disable=dangerous-default-value


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def compute_for_mcview(
    what: Union[str, ut.Matrix] = "__x__",
    *,
    adata: AnnData,
    gdata: AnnData,
    group: Union[str, ut.Vector] = "metacell",
    find_metacells_marker_genes: Optional[Dict[str, Any]] = {},
    compute_umap_by_markers_2: Optional[Dict[str, Any]] = {},
    compute_umap_by_markers_3: Optional[Dict[str, Any]] = {},
    compute_outliers_matches: Optional[Dict[str, Any]] = {},
    compute_deviant_folds: Optional[Dict[str, Any]] = {},
    compute_inner_folds: Optional[Dict[str, Any]] = {},
    count_significant_inner_folds: Optional[Dict[str, Any]] = {},
    compute_stdev_logs: Optional[Dict[str, Any]] = {},
    compute_var_var_similarity: Optional[Dict[str, Any]] = {"top": 50, "bottom": 50},
    random_seed: int,
) -> None:
    """
    Compute metacell analysis in preparation for exporting the data to MCView.

    This simply invokes a series of tools (which you can also invoke independently), using the default parameters,
    except for the ``group`` (default: {group}) which allows applying this to any clustering (not only metacells) and
    the ``random_seed`` (non-zero for reproducibility).

    If specific tool parameters need to be specified, you can pass them as a dictionary using the specific tool name
    (e.g., ``compute_umap_by_markers_2 = dict(spread = 0.5)``. If this parameter is set to ``None``, running the tool
    is skipped.

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes, where
    ``what`` is a per-variable-per-observation matrix or the name of a per-variable-per-observation
    annotation containing such a matrix.

    In addition, ``gdata`` is assumed to have one observation for each group, and use the same
    genes as ``adata``.

    **Returns**

    Sets many properties in ``gdata`` and some in ``adata``, see below.

    **Computation Parameters**

    0. Use the default parameters for each tool, unless a parameter with the same name provides specific parameter
       overrides for it. In addition, the special parameters ``group`` (default: {group}) and ``random_seed``,
       and the ``reproducible`` flag derived from it (true if the seed is not zero) are automatically
       passed to all relevant tools.

    1. Compute the "marker" metacell genes using :py:func:`metacells.tools.high.find_metacells_marker_genes`.

    2. Computes UMAP projections by invoking :py:func:`metacells.pipeline.umap.compute_umap_by_markers`. This is done
       twice, once with ``dimensions=2`` for visualization and once with ``dimensions=3`` to capture more of the
       manifold structure (used to automatically generate cluster colors). Therefore in this case there are two
       dictionary parameters ``compute_umap_by_markers_2`` and ``compute_umap_by_markers_3``.

    3. Compute for each outlier cell the "most similar" metacecell for it using
       :py:func:`metacells.tools.quality.compute_outliers_matches`.

    4. Compute for each gene and cell the fold factor between the gene's expression and that of the metacell
       it belongs to (or, for outliers, the one it is most similar to) using
       :py:func:`metacells.tools.quality.compute_deviant_folds`.

    5. Compute for each gene for each metacell the maximal of the above using
       :py:func:`metacells.tools.quality.compute_inner_folds`.

    6. Compute for each gene the number of metacells where the above is high using
       :py:func:`metacells.tools.quality.count_significant_inner_folds`.

    7. Compute for each gene for each metacell the standard deviation of the log (base 2) of the fractions of each gene
       across the cells of the metacell using :py:func:`metacells.tools.quality.compute_stdev_logs`.

    8. Compute the gene-gene (variable-variable) similarity matrix. Note by default this will use
       {compute_var_var_similarity} which aren't the normal defaults for ``compute_var_var_similarity``, in order to
       keep just the top correlated genes and bottom (anti-)correlated genes for each gene. Otherwise you will get a
       dense matrix of ~X0K by ~X0K entries, which typically isn't what you want.
    """
    reproducible = random_seed != 0

    ut.set_m_data(gdata, "mcview_format", "1.0")

    metacell_of_cells = ut.get_o_numpy(adata, group)
    ut.set_m_data(gdata, "outliers", np.sum(metacell_of_cells < 0))

    step_is_not_none = [
        find_metacells_marker_genes is not None,
        compute_umap_by_markers_3 is not None,
        compute_umap_by_markers_2 is not None,
        compute_outliers_matches is not None,
        compute_deviant_folds is not None,
        compute_inner_folds is not None,
        count_significant_inner_folds is not None,
        compute_stdev_logs is not None,
        compute_var_var_similarity is not None,
    ]
    steps_count = len([is_not_none for is_not_none in step_is_not_none if is_not_none])

    step_time = 1 / steps_count if ut.has_progress_bar() else None

    if find_metacells_marker_genes is not None:
        with ut.progress_bar_slice(step_time):
            tl.find_metacells_marker_genes(gdata, what, **find_metacells_marker_genes)

    if compute_umap_by_markers_3 is not None:
        with ut.progress_bar_slice(step_time):
            compute_umap_by_markers(gdata, what, dimensions=3, random_seed=random_seed, **compute_umap_by_markers_3)

    if compute_umap_by_markers_2 is not None:
        with ut.progress_bar_slice(step_time):
            compute_umap_by_markers(gdata, what, dimensions=2, random_seed=random_seed, **compute_umap_by_markers_2)

    if compute_outliers_matches is not None:
        with ut.progress_bar_slice(step_time):
            tl.compute_outliers_matches(
                what, adata=adata, gdata=gdata, group=group, reproducible=reproducible, **compute_outliers_matches
            )

    if compute_deviant_folds is not None:
        with ut.progress_bar_slice(step_time):
            tl.compute_deviant_folds(what, adata=adata, gdata=gdata, group=group, **compute_deviant_folds)

    if compute_inner_folds is not None:
        with ut.progress_bar_slice(step_time):
            tl.compute_inner_folds(adata=adata, gdata=gdata, group=group)

    if count_significant_inner_folds is not None:
        with ut.progress_bar_slice(step_time):
            tl.count_significant_inner_folds(gdata, **count_significant_inner_folds)

    if compute_stdev_logs is not None:
        with ut.progress_bar_slice(step_time):
            tl.compute_stdev_logs(what, adata=adata, gdata=gdata, group=group, **compute_stdev_logs)

    if compute_var_var_similarity is not None:
        with ut.progress_bar_slice(step_time):
            tl.compute_var_var_similarity(gdata, what, reproducible=reproducible, **compute_var_var_similarity)



"""
Cross-Similarity
----------------
"""

from typing import Optional
from typing import Union

import numpy as np
from anndata import AnnData  # type: ignore

import metacells.parameters as pr
import metacells.utilities as ut

__all__ = [
    "compute_obs_obs_similarity",
    "compute_var_var_similarity",
]


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def compute_obs_obs_similarity(
    adata: AnnData,
    what: Union[str, ut.Matrix] = "__x__",
    *,
    method: str = pr.similarity_method,
    logistics_location: float = pr.logistics_location,
    logistics_slope: float = pr.logistics_slope,
    top: Optional[int] = None,
    bottom: Optional[int] = None,
    inplace: bool = True,
    reproducible: bool,
) -> Optional[ut.PandasFrame]:
    """
    Compute a measure of the similarity between the observations (cells) of ``what`` (default: {what}).

    If ``reproducible`` is ``True``, a slower (still parallel) but
    reproducible algorithm will be used to compute Pearson correlations.

    The ``method`` (default: {method}) can be one of:
    * ``pearson`` for computing Pearson correlation.
    * ``abs_pearson`` for computing the absolute Pearson correlation.
    * ``repeated_pearson`` for computing correlations-of-correlations.
    * ``repeated_abs_pearson`` for computing absolute correlations-of-correlations.
    * ``logistics`` for computing the logistics function.
    * ``logistics_pearson`` for computing correlations-of-logistics.
    * ``logistics_abs_pearson`` for computing absolute correlations-of-logistics.

    If using the logistics function, use the ``logistics_slope`` (default: {logistics_slope}) and
    ``logistics_location`` (default: {logistics_location}).

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes, where
    ``what`` is a per-variable-per-observation matrix or the name of a per-variable-per-observation
    annotation containing such a matrix.

    **Returns**

    Observations-Pair (cells) Annotations
        ``obs_similarity``
            A square matrix where each entry is the similarity between a pair of cells.

    If ``inplace`` (default: {inplace}), this is written to the data, and the function returns
    ``None``. Otherwise this is returned as a pandas data frame (indexed by the observation names).

    **Computation Parameters**

    1. If ``method`` (default: {method}) is ``logistics`` or ``logistics_pearson``, compute the mean
       value of the logistics function between the variables of each pair of observations (cells).
       Otherwise, it should be ``pearson`` or ``repeated_pearson``, so compute the cross-correlation
       between all the observations.

    2. If the ``method`` is ``logistics_pearson`` or ``repeated_pearson``, then compute the
       cross-correlation of the results of the previous step. That is, two observations (cells) will
       be similar if they are similar to the rest of the observations (cells) in the same way. This
       compensates for the extreme sparsity of the data.

    3. If ``top`` and/or ``bottom`` are specified, keep just these number of most-similar and/or least-similar values in
       each row (turning the result into a compressed matrix format).
    """
    return _compute_elements_similarity(
        adata,
        "obs",
        "row",
        what,
        method=method,
        reproducible=reproducible,
        logistics_location=logistics_location,
        logistics_slope=logistics_slope,
        top=top,
        bottom=bottom,
        inplace=inplace,
    )


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def compute_var_var_similarity(
    adata: AnnData,
    what: Union[str, ut.Matrix] = "__x__",
    *,
    method: str = pr.similarity_method,
    logistics_location: float = pr.logistics_location,
    logistics_slope: float = pr.logistics_slope,
    top: Optional[int] = None,
    bottom: Optional[int] = None,
    inplace: bool = True,
    reproducible: bool,
) -> Optional[ut.PandasFrame]:
    """
    Compute a measure of the similarity between the variables (genes) of ``what`` (default: {what}).

    If ``reproducible`` is ``True``, a slower (still parallel) but
    reproducible algorithm will be used to compute Pearson correlations.

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes, where
    ``what`` is a per-variable-per-observation matrix or the name of a per-variable-per-observation
    annotation containing such a matrix.

    The ``method`` (default: {method}) can be one of:
    * ``pearson`` for computing Pearson correlation.
    * ``abs_pearson`` for computing the absolute Pearson correlation.
    * ``repeated_pearson`` for computing correlations-of-correlations.
    * ``repeated_abs_pearson`` for computing absolute correlations-of-correlations.
    * ``logistics`` for computing the logistics function.
    * ``logistics_pearson`` for computing correlations-of-logistics.
    * ``logistics_abs_pearson`` for computing absolute correlations-of-logistics.

    If using the logistics function, use the ``logistics_slope`` (default: {logistics_slope}) and
    ``logistics_location`` (default: {logistics_location}).

    **Returns**

    Variable-Pair (genes) Annotations
        ``var_similarity``
            A square matrix where each entry is the similarity between a pair of genes.

    If ``inplace`` (default: {inplace}), this is written to the data, and the function returns
    ``None``. Otherwise this is returned as a pandas data frame (indexed by the variable names).

    **Computation Parameters**

    1. If ``method`` (default: {method}) is ``logistics`` or ``logistics_pearson``, compute the mean
       value of the logistics function between the variables of each pair of variables (genes).
       Otherwise, it should be ``pearson`` or ``repeated_pearson``, so compute the cross-correlation
       between all the variables.

    2. If the ``method`` is ``logistics_pearson`` or ``repeated_pearson``, then compute the
       cross-correlation of the results of the previous step. That is, two variables (genes) will
       be similar if they are similar to the rest of the variables (genes) in the same way. This
       compensates for the extreme sparsity of the data.

    3. If ``top`` and/or ``bottom`` are specified, keep just these number of most-similar and/or least-similar values in
       each row (turning the result into a compressed matrix format).
    """
    return _compute_elements_similarity(
        adata,
        "var",
        "column",
        what,
        method=method,
        reproducible=reproducible,
        logistics_location=logistics_location,
        logistics_slope=logistics_slope,
        top=top,
        bottom=bottom,
        inplace=inplace,
    )


def _compute_elements_similarity(  # pylint: disable=too-many-branches
    adata: AnnData,
    elements: str,
    per: str,
    what: Union[str, ut.Matrix],
    *,
    method: str,
    reproducible: bool,
    logistics_location: float,
    logistics_slope: float,
    top: Optional[int],
    bottom: Optional[int],
    inplace: bool,
) -> Optional[ut.PandasFrame]:
    assert elements in ("obs", "var")

    assert method in (
        "logistics",
        "logistics_pearson",
        "logistics_abs_pearson",
        "pearson",
        "abs_pearson",
        "repeated_pearson",
        "repeated_abs_pearson",
    ), f"invalid similarity method: {method}"

    data = ut.get_vo_proper(adata, what, layout=f"{per}_major")
    dense = ut.to_numpy_matrix(data)

    similarity: ut.ProperMatrix
    if method in ("logistics", "logistics_pearson", "logistics_abs_pearson"):
        similarity = ut.logistics(dense, location=logistics_location, slope=logistics_slope, per=per)
        similarity *= -1
        similarity += 1
    else:
        similarity = ut.corrcoef(dense, per=per, reproducible=reproducible)
        if method in ("abs_pearson", "repeated_abs_pearson"):
            np.absolute(similarity, out=similarity)

    if method in ("repeated_pearson", "repeated_abs_pearson", "logistics_pearson", "logistics_abs_pearson"):
        similarity = ut.corrcoef(similarity, per=None, reproducible=reproducible)
        if method in ("repeated_abs_pearson", "logistics_abs_pearson"):
            np.absolute(similarity, out=similarity)

    if top is not None:
        top_similarity = ut.top_per(similarity, top, per="row")

    if bottom is not None:
        similarity *= -1
        bottom_similarity = ut.top_per(similarity, bottom, per="row")
        bottom_similarity *= -1  # type: ignore

    if top is not None:
        if bottom is not None:
            assert top + bottom <= similarity.shape[0]
            similarity = top_similarity + bottom_similarity  # type: ignore
        else:
            similarity = top_similarity
    else:
        if bottom is not None:
            similarity = bottom_similarity

    if inplace:
        to = elements + "_similarity"
        if elements == "obs":
            ut.set_oo_data(adata, to, similarity)
        else:
            ut.set_vv_data(adata, to, similarity)
        return None

    if elements == "obs":
        names = adata.obs_names
    else:
        names = adata.var_names

    return ut.to_pandas_frame(similarity, index=names, columns=names)



"""
Candidates
----------
"""

import sys
from math import ceil
from math import floor
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple
from typing import Union

import numpy as np
from anndata import AnnData  # type: ignore

import metacells.parameters as pr
import metacells.utilities as ut

if "sphinx" not in sys.argv[0]:
    import metacells.extensions as xt  # type: ignore

__all__ = [
    "compute_candidate_metacells",
    "choose_seeds",
    "optimize_partitions",
    "score_partitions",
]


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def compute_candidate_metacells(  # pylint: disable=too-many-statements
    adata: AnnData,
    what: Union[str, ut.Matrix] = "obs_outgoing_weights",
    *,
    target_metacell_size: int = pr.target_metacell_size,
    min_metacell_size: int = pr.min_metacell_size,
    target_metacell_umis: int = pr.target_metacell_umis,
    cell_umis: Optional[ut.NumpyVector] = pr.cell_umis,
    min_seed_size_quantile: float = pr.min_seed_size_quantile,
    max_seed_size_quantile: float = pr.max_seed_size_quantile,
    cooldown_pass: float = pr.cooldown_pass,
    cooldown_node: float = pr.cooldown_node,
    cooldown_phase: float = pr.cooldown_phase,
    increase_phase: float = 1.01,
    min_split_size_factor: float = pr.candidates_min_split_size_factor,
    max_merge_size_factor: float = pr.candidates_max_merge_size_factor,
    max_split_min_cut_strength: float = pr.max_split_min_cut_strength,
    min_cut_seed_cells: int = pr.min_cut_seed_cells,
    must_complete_cover: bool = False,
    random_seed: int,
    inplace: bool = True,
) -> Optional[ut.PandasSeries]:
    """
    Assign observations (cells) to (raw, candidate) metacells based on ``what`` data. (a weighted
    directed graph).

    These candidate metacells typically go through additional vetting (e.g. deviant detection and
    dissolving too-small metacells) to obtain the final metacells.

    **Input**

    Annotated ``adata``, where the observations are cells and the variables are genes, where
    ``what`` is a per-observation-per-observation matrix where each row is the outgoing weights from
    each observation to the rest, or just the name of a per-observation-per-observation annotation
    containing such a matrix. Typically this matrix will be sparse for efficient processing.

    **Returns**

    Observation (Cell) Annotations
        ``candidate``
            The integer index of the (raw, candidate) metacell each cell belongs to. The metacells
            are in no particular order.

    If ``inplace`` (default: {inplace}), this is written to the data, and the function returns
    ``None``. Otherwise this is returned as a pandas series (indexed by the variable names).

    **Computation Parameters**

    0. If ``cell_umis`` is not specified, use the sum of the ``what`` data for each cell.

    1. We are trying to create metacells of size ``target_metacell_size`` cells and ``target_metacell_umis`` UMIs each.
       Compute the UMIs of the metacells by summing the ``cell_umis``.

    2. We start with some an assignment of cells to seeds using :py:func:`choose_seeds` using ``min_seed_size_quantile``
       (default: {min_seed_size_quantile}) and ``max_seed_size_quantile`` (default: {max_seed_size_quantile}) to compute
       them, picking a number of seeds such that the average metacell size would match the target.

    3. We optimize the seeds using :py:func:`optimize_partitions` to obtain initial communities by
       maximizing the "stability" of the solution (probability of starting at a random node and
       moving either forward or backward in the graph and staying within the same metacell, divided
       by the probability of staying in the metacell if the edges connected random nodes). We pass
       it the ``cooldown_pass`` {cooldown_pass}) and ``cooldown_node`` (default: {cooldown_node}).

    4. If ``min_split_size_factor`` (default: {min_split_size_factor}) is specified, split to two each community whose
       size is partition method on each community whose size is at least
       ``target_metacell_size * min_split_size_factor`` or whose UMIs are at least
       ``target_metacell_umis * min_split_size_factor``, as long as half of the community is at least the
       ``min_metacell_size`` (default: {min_metacell_size}). Then, re-optimize the solution (resulting in an additional
       metacells). Every time we re-optimize, we multiply 1 - ``cooldown_pass`` by 1 - ``cooldown_phase`` (default:
       {cooldown_phase}).

    5. Using ``max_split_min_cut_strength`` (default: {max_split_min_cut_strength}), if
       the minimal cut of a candidate is lower, split it into two. If one of the partitions is
       smaller than ``min_cut_seed_cells``, then mark the cells in it as outliers, or if
       ``must_complete_cover`` is ``True``, skip the cut altogether.

    6. Using ``max_merge_size_factor`` (default: {max_merge_size_factor}) and ``min_metacell_size``
       (default: {min_metacell_size}), make outliers of cells of a community whose size is at most
       ``target_metacell_size * max_merge_size_factor`` and whose UMIs are at most
       ``target_metacell_umis * max_merge_size_factor``, or that contain less cells than ``min_metacell_size``. Again,
       re-optimize, which will assign these cells to other metacells (resulting on one less metacell). We again apply
       the ``cooldown_phase`` every time we re-optimize.

    7. Repeat the above steps until all metacells candidates are in the acceptable size range.
    """
    assert 0.0 < max_merge_size_factor < min_split_size_factor
    assert target_metacell_size > 0
    assert target_metacell_umis > 0
    assert 0.0 < cooldown_pass < 1.0
    assert 0.0 <= cooldown_node <= 1.0
    assert 0.0 < cooldown_phase <= 1.0

    edge_weights = ut.get_oo_proper(adata, what, layout="row_major")
    assert edge_weights.shape[0] == edge_weights.shape[1]
    size = edge_weights.shape[0]

    outgoing_edge_weights = ut.mustbe_compressed_matrix(edge_weights)

    assert ut.is_layout(outgoing_edge_weights, "row_major")
    incoming_edge_weights = ut.mustbe_compressed_matrix(ut.to_layout(outgoing_edge_weights, layout="column_major"))
    assert ut.is_layout(incoming_edge_weights, "column_major")

    assert outgoing_edge_weights.data.dtype == "float32"
    assert outgoing_edge_weights.indices.dtype == "int32"
    assert outgoing_edge_weights.indptr.dtype == "int32"
    assert incoming_edge_weights.data.dtype == "float32"
    assert incoming_edge_weights.indices.dtype == "int32"
    assert incoming_edge_weights.indptr.dtype == "int32"

    if cell_umis is None:
        cell_umis = ut.get_o_numpy(adata, what, sum=True, formatter=ut.sizes_description).astype("float32")
    else:
        assert cell_umis.dtype == "float32"
    assert isinstance(cell_umis, ut.NumpyVector)

    lowest_metacell_size = min_metacell_size
    min_metacell_umis = int(floor(target_metacell_umis * max_merge_size_factor))
    ut.log_calc("min_metacell_umis", min_metacell_umis)
    max_metacell_umis = int(ceil(target_metacell_umis * min_split_size_factor)) - 1
    ut.log_calc("max_metacell_umis", max_metacell_umis)

    min_metacell_size = max(min_metacell_size, int(floor(target_metacell_size * max_merge_size_factor)))
    ut.log_calc("min_metacell_size", min_metacell_size)
    max_metacell_size = max(int(ceil(target_metacell_size * min_split_size_factor)) - 1, min_metacell_size + 1)
    ut.log_calc("max_metacell_size", max_metacell_size)

    total_umis = sum(cell_umis)
    ut.log_calc("total_umis", total_umis)

    initial_seeds_count = _seeds_count_for(
        total_size=size,
        total_umis=total_umis,
        min_metacell_size=min_metacell_size,
        max_metacell_size=max_metacell_size,
        min_metacell_umis=min_metacell_umis,
        max_metacell_umis=max_metacell_umis,
    )

    community_of_nodes = np.full(size, -1, dtype="int32")
    initial_seeds_count = _choose_seeds(
        outgoing_edge_weights=outgoing_edge_weights,
        incoming_edge_weights=incoming_edge_weights,
        seed_of_cells=community_of_nodes,
        max_seeds_count=initial_seeds_count,
        min_seed_size_quantile=min_seed_size_quantile,
        max_seed_size_quantile=max_seed_size_quantile,
        random_seed=random_seed,
    )

    ut.set_o_data(adata, "seed", community_of_nodes, formatter=ut.groups_description)
    community_of_nodes = community_of_nodes.copy()

    np.random.seed(random_seed)

    cold_temperature = 1 - cooldown_pass

    old_score = 1e9
    old_communities = community_of_nodes
    old_small_nodes_count = len(community_of_nodes)
    atomic_candidates: Set[Tuple[int, ...]] = set()
    hot_communities = list(range(initial_seeds_count))

    while True:
        cold_temperature, score = _reduce_communities(
            outgoing_edge_weights=outgoing_edge_weights,
            incoming_edge_weights=incoming_edge_weights,
            community_of_nodes=community_of_nodes,
            hot_communities=hot_communities,
            target_metacell_umis=target_metacell_umis,
            node_umis=cell_umis,
            min_metacell_umis=min_metacell_umis,
            max_metacell_umis=max_metacell_umis,
            target_metacell_size=target_metacell_size,
            min_metacell_size=min_metacell_size,
            max_metacell_size=max_metacell_size,
            max_split_min_cut_strength=max_split_min_cut_strength,
            min_cut_seed_cells=min_cut_seed_cells,
            must_complete_cover=must_complete_cover,
            min_seed_size_quantile=min_seed_size_quantile,
            max_seed_size_quantile=max_seed_size_quantile,
            random_seed=random_seed,
            cooldown_pass=cooldown_pass,
            cooldown_node=cooldown_node,
            cooldown_phase=cooldown_phase,
            increase_phase=increase_phase,
            cold_temperature=cold_temperature,
            atomic_candidates=atomic_candidates,
        )

        small_communities, small_nodes_count, small_nodes_umis = _find_small_communities(
            community_of_nodes=community_of_nodes,
            node_umis=cell_umis,
            min_metacell_umis=min_metacell_umis,
            max_metacell_umis=max_metacell_umis,
            min_metacell_size=min_metacell_size,
            max_metacell_size=max_metacell_size,
            lowest_metacell_size=lowest_metacell_size,
        )

        small_communities_count = len(small_communities)
        ut.log_calc("small communities count", small_communities_count)
        if small_communities_count < 2:
            break

        if (old_small_nodes_count, old_score) <= (small_nodes_count, score):
            ut.logger().debug("is not better, revert")
            community_of_nodes = old_communities
            score = old_score
            ut.log_calc("communities", community_of_nodes, formatter=ut.groups_description)
            ut.log_calc("score", score)
            break

        old_score = score
        old_communities = community_of_nodes.copy()
        old_small_nodes_count = small_nodes_count

        kept_communities_count = _cancel_communities(
            community_of_nodes=community_of_nodes, cancelled_communities=small_communities
        )

        small_communities_seeds = min(
            small_communities_count - 1,
            _seeds_count_for(
                total_size=small_nodes_count,
                total_umis=small_nodes_umis,
                min_metacell_size=min_metacell_size,
                max_metacell_size=max_metacell_size,
                min_metacell_umis=min_metacell_umis,
                max_metacell_umis=max_metacell_umis,
            ),
        )
        ut.log_calc("small communities seeds", small_communities_seeds)

        max_seeds_count = kept_communities_count + small_communities_seeds
        seeds_count = _choose_seeds(
            outgoing_edge_weights=outgoing_edge_weights,
            incoming_edge_weights=incoming_edge_weights,
            seed_of_cells=community_of_nodes,
            max_seeds_count=max_seeds_count,
            min_seed_size_quantile=min_seed_size_quantile,
            max_seed_size_quantile=max_seed_size_quantile,
            random_seed=random_seed,
        )

        hot_communities = list(range(kept_communities_count, seeds_count))

    assert np.all(community_of_nodes >= 0)

    if inplace:
        ut.set_o_data(adata, "candidate", community_of_nodes, formatter=ut.groups_description)
        return None

    ut.log_return("candidate", community_of_nodes, formatter=ut.groups_description)
    return ut.to_pandas_series(community_of_nodes, index=adata.obs_names)


@ut.logged()
def _seeds_count_for(
    *,
    total_umis: float,
    total_size: float,
    min_metacell_size: float,
    max_metacell_size: float,
    min_metacell_umis: float,
    max_metacell_umis: float,
) -> int:
    min_seeds_count_by_umis = int(ceil(total_umis / max_metacell_umis))
    ut.log_calc("min_seeds_count_by_umis", min_seeds_count_by_umis)
    max_seeds_count_by_umis = int(ceil(total_umis / min_metacell_umis))
    ut.log_calc("max_seeds_count_by_umis", max_seeds_count_by_umis)

    min_seeds_count_by_size = int(ceil(total_size / max_metacell_size))
    ut.log_calc("min_seeds_count_by_size", min_seeds_count_by_size)
    max_seeds_count_by_size = int(ceil(total_size / min_metacell_size))
    ut.log_calc("max_seeds_count_by_size", max_seeds_count_by_size)

    if max_seeds_count_by_size < min_seeds_count_by_umis:
        seeds_count = int(round((max_seeds_count_by_size + min_seeds_count_by_umis) / 2))
    elif max_seeds_count_by_umis < min_seeds_count_by_size:
        seeds_count = int(round((max_seeds_count_by_umis + min_seeds_count_by_size) / 2))
    else:
        min_seeds_count = max(min_seeds_count_by_size, min_seeds_count_by_umis)
        max_seeds_count = min(max_seeds_count_by_size, max_seeds_count_by_umis)
        seeds_count = int(round((min_seeds_count + max_seeds_count) / 2))
    ut.log_return("seeds_count", seeds_count)
    return seeds_count


def _reduce_communities(
    outgoing_edge_weights: ut.CompressedMatrix,
    incoming_edge_weights: ut.CompressedMatrix,
    community_of_nodes: ut.NumpyVector,
    hot_communities: List[int],
    target_metacell_umis: float,
    node_umis: ut.NumpyVector,
    min_metacell_umis: float,
    max_metacell_umis: float,
    target_metacell_size: float,
    min_metacell_size: float,
    max_metacell_size: float,
    max_split_min_cut_strength: float,
    min_cut_seed_cells: int,
    must_complete_cover: bool,
    min_seed_size_quantile: float,
    max_seed_size_quantile: float,
    random_seed: int,
    cooldown_pass: float,
    cooldown_node: float,
    cooldown_phase: float,
    increase_phase: float,
    cold_temperature: float,
    atomic_candidates: Set[Tuple[int, ...]],
) -> Tuple[float, float]:
    np.random.seed(random_seed)
    while True:
        ut.log_calc("cold_temperature", cold_temperature)
        score = _optimize_partitions(
            outgoing_edge_weights=outgoing_edge_weights,
            incoming_edge_weights=incoming_edge_weights,
            community_of_nodes=community_of_nodes,
            hot_communities=hot_communities,
            node_umis=node_umis,
            low_partition_umis=min_metacell_umis,
            target_partition_umis=target_metacell_umis,
            high_partition_umis=max_metacell_umis,
            low_partition_size=min_metacell_size,
            target_partition_size=target_metacell_size,
            high_partition_size=max_metacell_size,
            random_seed=random_seed,
            cooldown_pass=cooldown_pass,
            cooldown_node=cooldown_node,
            cold_temperature=cold_temperature,
        )

        cold_temperature = cold_temperature * (1 - cooldown_phase)
        max_metacell_size *= increase_phase
        ut.log_calc("communities", community_of_nodes, formatter=ut.groups_description)
        ut.log_calc("max_metacell_size", max_metacell_size)

        split_communities_count, cut_communities_count, hot_communities = _cut_split_communities(
            outgoing_edge_weights=outgoing_edge_weights,
            community_of_nodes=community_of_nodes,
            target_metacell_umis=target_metacell_umis,
            node_umis=node_umis,
            min_metacell_umis=min_metacell_umis,
            max_metacell_umis=max_metacell_umis,
            target_metacell_size=target_metacell_size,
            min_metacell_size=min_metacell_size,
            max_metacell_size=max_metacell_size,
            max_split_min_cut_strength=max_split_min_cut_strength,
            min_cut_seed_cells=min_cut_seed_cells,
            must_complete_cover=must_complete_cover,
            atomic_candidates=atomic_candidates,
        )

        if split_communities_count + cut_communities_count == 0:
            return cold_temperature, score

        max_seeds_count = np.max(community_of_nodes) + 1
        if cut_communities_count > 0:
            hot_communities.append(max_seeds_count)
            max_seeds_count += 1

        _choose_seeds(
            outgoing_edge_weights=outgoing_edge_weights,
            incoming_edge_weights=incoming_edge_weights,
            seed_of_cells=community_of_nodes,
            max_seeds_count=max_seeds_count,
            min_seed_size_quantile=min_seed_size_quantile,
            max_seed_size_quantile=max_seed_size_quantile,
            random_seed=random_seed,
        )


def _cut_split_communities(  # pylint: disable=too-many-branches,too-many-statements
    *,
    outgoing_edge_weights: ut.CompressedMatrix,
    community_of_nodes: ut.NumpyVector,
    target_metacell_umis: float,
    node_umis: ut.NumpyVector,
    min_metacell_umis: float,
    max_metacell_umis: float,
    target_metacell_size: float,
    min_metacell_size: float,
    max_metacell_size: float,
    max_split_min_cut_strength: float,
    min_cut_seed_cells: int,
    must_complete_cover: bool,
    atomic_candidates: Set[Tuple[int, ...]],
) -> Tuple[int, int, List[int]]:
    communities_count = np.max(community_of_nodes) + 1
    assert communities_count > 0

    split_communities_count = 0
    cut_communities_count = 0
    next_new_community_index = communities_count

    hot_communities = []

    for community_index in range(communities_count):
        community_mask = community_of_nodes == community_index
        community_size = np.sum(community_mask)
        community_umis = np.sum(node_umis[community_mask])

        if (community_umis > max_metacell_umis and community_size >= 3 * min_metacell_size) or (
            community_size > max_metacell_size and community_umis >= 3 * min_metacell_umis
        ):
            community_indices = np.where(community_mask)[0]
            split_parts_count = max(
                2,
                min(
                    int(round(community_umis / target_metacell_umis)), int(round(community_size / target_metacell_size))
                ),
            )
            ut.logger().debug(
                "community: %s size: %s umis: %s is too large, split into %s",
                community_index,
                community_size,
                community_umis,
                split_parts_count,
            )
            all_parts_exist = False
            while not all_parts_exist:
                split_parts_assignment = np.random.randint(  # type: ignore
                    split_parts_count, size=len(community_indices)
                )
                all_parts_exist = True
                for split_index in range(split_parts_count):
                    if not np.any(split_parts_assignment == split_index):
                        all_parts_exist = False
                        break
            community_of_nodes[community_indices] = -1
            for split_index in range(split_parts_count):
                split_part_mask = split_parts_assignment == split_index
                assert np.any(split_part_mask)
                split_part_indices = community_indices[split_part_mask]
                if split_index == 0:
                    hot_communities.append(community_index)
                    community_of_nodes[split_part_indices[np.random.randint(len(split_part_indices))]] = community_index
                else:
                    hot_communities.append(next_new_community_index)
                    community_of_nodes[
                        split_part_indices[np.random.randint(len(split_part_indices))]
                    ] = next_new_community_index
                    next_new_community_index += 1
                    split_communities_count += 1
            continue

    if split_communities_count == 0:
        for community_index in range(communities_count):
            community_mask = community_of_nodes == community_index
            community_indices_key = tuple(np.where(community_mask)[0])
            if community_indices_key in atomic_candidates:
                continue

            action = _min_cut_community(
                outgoing_edge_weights=outgoing_edge_weights,
                community_of_nodes=community_of_nodes,
                cut_community_index=community_index,
                max_split_min_cut_strength=max_split_min_cut_strength,
                min_cut_seed_cells=min_cut_seed_cells,
                must_complete_cover=must_complete_cover,
                new_community_index=next_new_community_index,
            )

            if action == "unchanged":
                atomic_candidates.add(community_indices_key)
                continue

            hot_communities.append(community_index)
            ut.logger().debug(
                "community: %s nodes: %s size: %s was %s",
                community_index,
                np.sum(community_mask),
                community_umis,
                action,
            )

            if action == "split":
                hot_communities.append(next_new_community_index)
                split_communities_count += 1
                next_new_community_index += 1
            else:
                cut_communities_count += 1

    if split_communities_count + cut_communities_count > 0:
        ut.logger().debug("old communities: %s", communities_count)
        ut.logger().debug("cut communities: %s", cut_communities_count)
        ut.logger().debug("split communities: %s", split_communities_count)
        ut.logger().debug("hot communities: %s", len(hot_communities))
        ut.logger().debug("total communities: %s", communities_count + split_communities_count)
    else:
        ut.logger().debug("no communities were cut or split")

    return (split_communities_count, cut_communities_count, hot_communities)


def _min_cut_community(
    outgoing_edge_weights: ut.CompressedMatrix,
    community_of_nodes: ut.NumpyVector,
    cut_community_index: int,
    max_split_min_cut_strength: float,
    min_cut_seed_cells: int,
    must_complete_cover: bool,
    new_community_index: int,
) -> str:
    community_mask = community_of_nodes == cut_community_index
    if np.sum(community_mask) < 2:
        return "unchanged"
    community_edge_weights = outgoing_edge_weights[community_mask, :][:, community_mask]
    community_edge_weights += community_edge_weights.T

    cut, cut_strength = ut.min_cut(community_edge_weights)
    if cut_strength is None:
        return "unchanged"

    if cut_strength == 0:
        community_indices = np.where(community_mask)[0]
        if len(cut.partition[0]) < len(cut.partition[1]):
            small_partition = 0
        else:
            small_partition = 1
        small_community_indices = community_indices[cut.partition[small_partition]]

    if cut_strength > max_split_min_cut_strength:
        return "unchanged"

    ut.logger().debug(
        "min cut community: %s partitions: %s + %s = %s strength: %s",
        cut_community_index,
        len(cut.partition[0]),
        len(cut.partition[1]),
        community_edge_weights.shape[0],
        cut_strength,
    )

    community_indices = np.where(community_mask)[0]
    if len(cut.partition[0]) < len(cut.partition[1]):
        small_partition = 0
    else:
        small_partition = 1

    if len(cut.partition[small_partition]) >= min_cut_seed_cells:
        second_partition_indices = community_indices[cut.partition[1]]
        community_of_nodes[second_partition_indices] = new_community_index
        return "split"

    if must_complete_cover and cut_strength > 0:
        ut.logger().debug("give up on small cut: %s", len(cut.partition[small_partition]))
        return "unchanged"

    small_community_indices = community_indices[cut.partition[small_partition]]
    community_of_nodes[small_community_indices] = -1
    return "cut"


def _find_small_communities(
    *,
    community_of_nodes: ut.NumpyVector,
    node_umis: ut.NumpyVector,
    min_metacell_umis: float,
    max_metacell_umis: float,
    min_metacell_size: float,
    max_metacell_size: float,
    lowest_metacell_size: float,
) -> Tuple[Set[int], int, float]:
    communities_count = np.max(community_of_nodes) + 1
    assert communities_count > 0

    small_communities: Set[int] = set()
    small_nodes_count = 0
    small_nodes_umis = 0

    for community_index in range(communities_count):
        community_mask = community_of_nodes == community_index
        community_size = np.sum(community_mask)
        community_umis = np.sum(node_umis[community_mask])

        if community_size < lowest_metacell_size:
            ut.logger().debug("community: %s size: %s is tiny", community_index, community_size)
        elif community_size < min_metacell_size and 2 * community_umis < max_metacell_umis:
            ut.logger().debug(
                "community: %s umis: %s and size: %s is too few", community_index, community_umis, community_size
            )
        elif community_umis < min_metacell_umis and 2 * community_size < max_metacell_size:
            ut.logger().debug(
                "community: %s umis: %s and size: %s is too small", community_index, community_umis, community_size
            )
        else:
            continue

        small_communities.add(community_index)
        small_nodes_count += community_size
        small_nodes_umis += community_umis

    return (small_communities, small_nodes_count, small_nodes_umis)


def _cancel_communities(community_of_nodes: ut.NumpyVector, cancelled_communities: Set[int]) -> int:
    communities_count = np.max(community_of_nodes) + 1
    kept_communities_count = 0
    for community_index in range(communities_count):
        if community_index in cancelled_communities:
            community_of_nodes[community_of_nodes == community_index] = -1
            continue

        if community_index > kept_communities_count:
            community_of_nodes[community_of_nodes == community_index] = kept_communities_count

        kept_communities_count += 1

    assert kept_communities_count == communities_count - len(cancelled_communities)
    return kept_communities_count


@ut.logged()
@ut.timed_call()
def choose_seeds(
    *,
    edge_weights: ut.CompressedMatrix,
    seed_of_cells: Optional[ut.NumpyVector] = None,
    max_seeds_count: int,
    min_seed_size_quantile: float = pr.min_seed_size_quantile,
    max_seed_size_quantile: float = pr.max_seed_size_quantile,
    random_seed: int,
) -> ut.NumpyVector:
    """
    Choose initial assignment of cells to seeds based on the ``edge_weights``.

    Returns a vector assigning each node (cell) to a seed (initial community).

    If ``seed_of_cells`` is specified, it is expected to contain a vector of partial seeds. Only
    cells which have a negative seed will be assigned a new seed. New seeds will be created so that
    the total number of seeds will not exceed ``max_seeds_count``. The ``seed_of_cells`` will be
    modified in-place and returned.

    Otherwise, a new vector is created, initialized with ``-1`` (that is, no seed) for all nodes,
    filled as above, and returned.

    **Computation Parameters**

    1. We compute for each candidate node the number of nodes it is connected to (by an outgoing
       edge).

    2. We pick as a seed a random node whose number of connected nodes ("seed size") quantile is at
       least ``min_seed_size_quantile`` and at most ``max_seed_size_quantile``. This ensures we pick
       seeds that aren't too small or too large to get a good coverage of the population with a low
       number of seeds.

    3. We assign each of the connected nodes to their seed, and discount them from the number of
       connected nodes of the remaining unassigned nodes.

    4. We repeat this until we reach the target number of seeds.
    """
    size = edge_weights.shape[0]

    outgoing_edge_weights = ut.mustbe_compressed_matrix(edge_weights)
    assert ut.is_layout(outgoing_edge_weights, "row_major")

    incoming_edge_weights = ut.mustbe_compressed_matrix(ut.to_layout(outgoing_edge_weights, layout="column_major"))
    assert ut.is_layout(incoming_edge_weights, "column_major")

    if seed_of_cells is None:
        seed_of_cells = np.full(size, -1, dtype="int32")
    else:
        assert seed_of_cells.dtype == "int32"

    assert outgoing_edge_weights.shape == incoming_edge_weights.shape == (len(seed_of_cells), len(seed_of_cells))

    _choose_seeds(
        outgoing_edge_weights=outgoing_edge_weights,
        incoming_edge_weights=incoming_edge_weights,
        seed_of_cells=seed_of_cells,
        max_seeds_count=max_seeds_count,
        min_seed_size_quantile=min_seed_size_quantile,
        max_seed_size_quantile=max_seed_size_quantile,
        random_seed=random_seed,
    )

    return seed_of_cells


@ut.logged()
@ut.timed_call()
def _choose_seeds(
    *,
    outgoing_edge_weights: ut.CompressedMatrix,
    incoming_edge_weights: ut.CompressedMatrix,
    seed_of_cells: ut.NumpyVector,
    max_seeds_count: int,
    min_seed_size_quantile: float,
    max_seed_size_quantile: float,
    random_seed: int,
) -> int:
    ut.log_calc("partial seeds", seed_of_cells, formatter=ut.groups_description)

    seeds_count = xt.choose_seeds(
        outgoing_edge_weights.data,
        outgoing_edge_weights.indices,
        outgoing_edge_weights.indptr,
        incoming_edge_weights.data,
        incoming_edge_weights.indices,
        incoming_edge_weights.indptr,
        random_seed,
        max_seeds_count,
        min_seed_size_quantile,
        max_seed_size_quantile,
        seed_of_cells,
    )

    ut.log_calc("chosen seeds count", seeds_count)
    ut.log_calc("chosen seeds", seed_of_cells, formatter=ut.groups_description)
    assert np.min(seed_of_cells) == 0
    return seeds_count


@ut.logged()
@ut.timed_call()
def optimize_partitions(
    *,
    edge_weights: ut.CompressedMatrix,
    community_of_nodes: ut.NumpyVector,
    node_umis: ut.NumpyVector,
    low_partition_umis: int,
    target_partition_umis: int,
    high_partition_umis: int,
    low_partition_size: int,
    target_partition_size: int,
    high_partition_size: int,
    cooldown_pass: float = pr.cooldown_pass,
    cooldown_node: float = pr.cooldown_node,
    random_seed: int,
) -> float:
    """
    Optimize partition to candidate metacells (communities) using the ``edge_weights``.

    Returns the score of the optimized partition.

    This modifies the ``community_of_nodes`` in-place.

    The goal is to minimize the "stability" goal function which is defined to be the ratio between
    (1) the probability that, selecting a random node and either a random outgoing edge or a random
    incoming edge (biased by their weights), that the node connected to by that edge is in the same
    community (metacell) and (2) the probability that a random edge would lead to this same
    community (the fraction of its number of nodes out of the total).

    To maximize this, we repeatedly pass on a randomized permutation of the nodes, and for each
    node, move it to a random "better" community. When deciding if a community is better, we
    consider both (1) just the "local" product of the sum of the weights of incoming and outgoing edges
    between the node and the current and candidate communities and (2) the effect on the "global" goal
    function (considering the impact on this product for all other nodes connected to the current
    node).

    We define a notion of ``temperature`` (initially, 1 - ``cooldown_pass``, default:
    {cooldown_pass}) and we give a weight of ``temperature`` to the local score and
    (1 - ``temperature``) to the global score. When we move to the next node, we multiply the
    temperature by 1 - ``cooldown_pass``. If we did not move the node, we multiply its temperature
    by ``cooldown_node`` (default: {cooldown_node}). We skip looking at nodes which are colder from
    the global temperature to accelerate the algorithm. If we don't move any node, we reduce the
    global temperature below that of any cold node; if there are no such nodes, we reduce it to zero
    to perform a final hill-climbing phase.

    This simulated-annealing-like behavior helps the algorithm to escape local maximums, although of
    course no claim is made of achieving the global maximum of the goal function.
    """
    outgoing_edge_weights = ut.mustbe_compressed_matrix(edge_weights)
    assert ut.is_layout(outgoing_edge_weights, "row_major")
    assert 0 < low_partition_size < target_partition_size < high_partition_size
    assert 0 < low_partition_umis < target_partition_umis < high_partition_umis

    incoming_edge_weights = ut.mustbe_compressed_matrix(ut.to_layout(outgoing_edge_weights, layout="column_major"))
    communities_count = np.max(community_of_nodes) + 1
    assert communities_count > 0
    assert ut.is_layout(incoming_edge_weights, "column_major")
    return _optimize_partitions(
        outgoing_edge_weights=outgoing_edge_weights,
        incoming_edge_weights=incoming_edge_weights,
        random_seed=random_seed,
        node_umis=node_umis,
        target_partition_umis=target_partition_umis,
        low_partition_umis=low_partition_umis,
        high_partition_umis=high_partition_umis,
        target_partition_size=target_partition_size,
        low_partition_size=low_partition_size,
        high_partition_size=high_partition_size,
        cooldown_pass=cooldown_pass,
        cooldown_node=cooldown_node,
        community_of_nodes=community_of_nodes,
        hot_communities=list(range(communities_count)),
        cold_temperature=cooldown_pass,
    )


@ut.logged()
@ut.timed_call()
def _optimize_partitions(
    *,
    outgoing_edge_weights: ut.CompressedMatrix,
    incoming_edge_weights: ut.CompressedMatrix,
    community_of_nodes: ut.NumpyVector,
    hot_communities: List[int],
    node_umis: ut.NumpyVector,
    low_partition_umis: float,
    target_partition_umis: float,
    high_partition_umis: float,
    target_partition_size: float,
    low_partition_size: float,
    high_partition_size: float,
    cooldown_pass: float,
    cooldown_node: float,
    cold_temperature: float,
    random_seed: int,
) -> float:
    assert community_of_nodes.dtype == "int32"
    assert node_umis.dtype == "float32"

    communities_count = np.max(community_of_nodes) + 1
    hot_communities_mask = np.zeros(max(communities_count, 2), dtype="int8")
    hot_communities_mask[hot_communities] = 1

    score = xt.optimize_partitions(
        outgoing_edge_weights.data,
        outgoing_edge_weights.indices,
        outgoing_edge_weights.indptr,
        incoming_edge_weights.data,
        incoming_edge_weights.indices,
        incoming_edge_weights.indptr,
        random_seed,
        node_umis,
        low_partition_umis,
        target_partition_umis,
        high_partition_umis,
        low_partition_size,
        target_partition_size,
        high_partition_size,
        cooldown_pass,
        cooldown_node,
        community_of_nodes,
        hot_communities_mask,
        cold_temperature,
    )
    ut.log_calc("score", score)
    ut.log_calc("partitions", community_of_nodes, formatter=ut.groups_description)
    assert np.min(community_of_nodes) == 0
    return score


@ut.logged()
@ut.timed_call()
def score_partitions(
    *,
    node_umis: ut.NumpyVector,
    low_partition_umis: float,
    target_partition_umis: float,
    high_partition_umis: float,
    low_partition_size: int,
    target_partition_size: int,
    high_partition_size: int,
    edge_weights: ut.CompressedMatrix,
    partition_of_nodes: ut.NumpyVector,
    temperature: float,
    with_orphans: bool = True,
) -> None:
    """
    Compute the "stability" the "stability" goal function which is defined to be the ratio between
    (1) the probability that, selecting a random node and either a random outgoing edge or a random
    incoming edge (biased by their weights), that the node connected to by that edge is in the same
    community (metacell) and (2) the probability that a random edge would lead to this same
    community (the fraction of its number of nodes out of the total).

    If ``with_orphans`` is True (the default), outlier nodes are included in the computation. In
    general we add 1e-6 to the product of the incoming and outgoing weights so we can safely log it
    for efficient computation; thus orphans are given a very small (non-zero) weight so the overall
    score is not zeroed even when including them.
    """
    assert partition_of_nodes.dtype == "int32"
    assert node_umis.dtype == "float32"
    outgoing_edge_weights = ut.mustbe_compressed_matrix(edge_weights)
    assert ut.is_layout(outgoing_edge_weights, "row_major")

    incoming_edge_weights = ut.mustbe_compressed_matrix(ut.to_layout(outgoing_edge_weights, layout="column_major"))
    assert ut.is_layout(incoming_edge_weights, "column_major")

    hot_communities_mask = np.zeros(np.max(partition_of_nodes) + 1, dtype="int8")

    with ut.unfrozen(partition_of_nodes):
        with ut.timed_step(".score"):
            score = xt.score_partitions(
                outgoing_edge_weights.data,
                outgoing_edge_weights.indices,
                outgoing_edge_weights.indptr,
                incoming_edge_weights.data,
                incoming_edge_weights.indices,
                incoming_edge_weights.indptr,
                node_umis,
                low_partition_umis,
                target_partition_umis,
                high_partition_umis,
                low_partition_size,
                target_partition_size,
                high_partition_size,
                temperature,
                partition_of_nodes,
                hot_communities_mask,
                with_orphans,
            )

    ut.log_calc("score", score)
    return score



"""
Named
-----
"""

from re import Pattern
from typing import Collection
from typing import Optional
from typing import Union

import numpy as np
from anndata import AnnData  # type: ignore

import metacells.utilities as ut

__all__ = [
    "find_named_genes",
]


@ut.logged()
@ut.timed_call()
@ut.expand_doc()
def find_named_genes(  # pylint: disable=too-many-branches
    adata: AnnData,
    *,
    name_property: Optional[str] = None,
    names: Optional[Collection[str]] = None,
    patterns: Optional[Collection[Union[str, Pattern]]] = None,
    to: Optional[str] = None,
    invert: bool = False,
    op: str = "set",
) -> Optional[ut.PandasSeries]:
    """
    Find genes by their (case-insensitive) name.

    This computes a mask of all the genes whose name appears in ``names`` or matches any of the ``patterns``. If
    ``invert`` (default: {invert}), invert the resulting mask.

    Depending on ``op``, this will ``set`` a (compute a brand new) mask, ``add`` the result to a mask (which must
    exist), or ``remove`` genes from a mask (which must exist).

    If ``name_property`` is specified the mask will be based on this per-variable (gene) property.

    If ``to`` (default: {to}) is specified, this is stored as a per-variable (gene) annotation with that name, and
    returns ``None``. This is useful to fill gene masks such as ``excluded_genes`` (genes which should be excluded from
    the rest of the processing), ``lateral_genes`` (genes which must not be selected for metacell computation) and
    ``noisy_genes`` (genes which are given more leeway when computing deviant cells).

    Otherwise, it returns it as a pandas series (indexed by the variable, that is gene, names).
    """
    assert op in ("set", "add", "remove")
    if op in ("add", "remove"):
        assert to is not None
        base_mask = ut.get_v_numpy(adata, to)
    else:
        base_mask = np.zeros(adata.n_vars, dtype="bool")

    if name_property is None:
        var_names = adata.var_names
    else:
        var_names = ut.get_v_numpy(adata, name_property)

    if names is None or len(names) == 0:
        names_mask = np.zeros(adata.n_vars, dtype="bool")
    else:
        lower_names_set = {name.lower() for name in names}
        names_mask = np.array([name.lower() in lower_names_set for name in var_names])  #

    if patterns is None or len(patterns) == 0:
        patterns_mask = np.zeros(adata.n_vars, dtype="bool")
    else:
        patterns_mask = ut.patterns_matches(patterns, var_names)

    genes_mask = names_mask | patterns_mask

    if invert:
        genes_mask = ~genes_mask

    if op == "add":
        result_mask = base_mask | genes_mask
    elif op == "remove":
        result_mask = base_mask & ~genes_mask
    else:
        assert op == "set"
        result_mask = genes_mask

    if to is not None:
        ut.set_v_data(adata, to, result_mask)
        return None

    ut.log_return("named_genes", result_mask)
    return ut.to_pandas_series(result_mask, index=adata.var_names)
